# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import tempfile
import os
import pickle
import random
import collections
import itertools
from io import StringIO

import pyomo.common.unittest as unittest
from pyomo.core.expr.numvalue import native_numeric_types
from pyomo.core.expr.symbol_map import SymbolMap
import pyomo.kernel as pmo
from pyomo.common.log import LoggingIntercept
from pyomo.core.tests.unit.kernel.test_dict_container import (
    _TestActiveDictContainerBase,
)
from pyomo.core.tests.unit.kernel.test_tuple_container import (
    _TestActiveTupleContainerBase,
)
from pyomo.core.tests.unit.kernel.test_list_container import (
    _TestActiveListContainerBase,
)
from pyomo.core.kernel.base import ICategorizedObject, ICategorizedObjectContainer
from pyomo.core.kernel.heterogeneous_container import (
    heterogeneous_containers,
    IHeterogeneousContainer,
)
from pyomo.common.collections import ComponentMap
from pyomo.core.kernel.suffix import suffix
from pyomo.core.kernel.constraint import (
    IConstraint,
    constraint,
    constraint_dict,
    constraint_list,
)
from pyomo.core.kernel.parameter import parameter, parameter_dict, parameter_list
from pyomo.core.kernel.expression import (
    expression,
    data_expression,
    expression_dict,
    expression_list,
)
from pyomo.core.kernel.objective import objective, objective_dict, objective_list
from pyomo.core.kernel.variable import IVariable, variable, variable_dict, variable_list
from pyomo.core.kernel.block import IBlock, block, block_dict, block_tuple, block_list
from pyomo.core.kernel.sos import sos
from pyomo.opt.results import Solution


def _path_to_object_exists(obj, descendent):
    if descendent is obj:
        return True
    else:
        parent = descendent.parent
        if parent is None:
            return False
        else:
            return _path_to_object_exists(obj, parent)


def _active_path_to_object_exists(obj, descendent):
    if descendent is obj:
        return True
    else:
        parent = descendent.parent
        if parent is None:
            return False
        else:
            if getattr(descendent, "active", True):
                return _active_path_to_object_exists(obj, parent)
            else:
                return False


def _collect_expr_components(exp):
    ans = {}
    if isinstance(exp, ICategorizedObject):
        ans[id(exp)] = exp
    if exp.__class__ in native_numeric_types:
        return ans
    if exp.is_expression_type():
        for subexp in exp.args:
            ans.update(_collect_expr_components(subexp))
    return ans


class IJunk(IBlock):
    __slots__ = ()


class junk(pmo.block):
    _ctype = IJunk


class junk_list(pmo.block_list):
    __slots__ = ()
    _ctype = IJunk


class TestHeterogeneousContainer(unittest.TestCase):
    model = pmo.block()
    model.v = pmo.variable()
    model.V = pmo.variable_list()
    model.V.append(pmo.variable())
    model.V.append(pmo.variable_list())
    model.V[1].append(pmo.variable())
    model.c = pmo.constraint()
    model.C = pmo.constraint_list()
    model.C.append(pmo.constraint())
    model.C.append(pmo.constraint_list())
    model.C[1].append(pmo.constraint())
    b_clone = model.clone()
    model.b = b_clone.clone()
    model.B = pmo.block_list()
    model.B.append(b_clone.clone())
    model.B.append(pmo.block_list())
    model.B[1].append(b_clone.clone())
    del b_clone
    model.j = junk()
    model.J = junk_list()
    model.J.append(junk())
    model.J.append(junk_list())
    model.J[1].append(junk())
    model.J[1][0].b = pmo.block()
    model.J[1][0].b.v = pmo.variable()
    model_clone = model.clone()
    model.k = pmo.block()
    model.K = pmo.block_list()
    model.K.append(model_clone.clone())
    del model_clone

    def test_preorder_traversal(self):
        model = self.model.clone()

        order = list(str(obj) for obj in pmo.preorder_traversal(model))
        self.assertEqual(
            order,
            [
                '<block>',
                'v',
                'V',
                'V[0]',
                'V[1]',
                'V[1][0]',
                'c',
                'C',
                'C[0]',
                'C[1]',
                'C[1][0]',
                'b',
                'b.v',
                'b.V',
                'b.V[0]',
                'b.V[1]',
                'b.V[1][0]',
                'b.c',
                'b.C',
                'b.C[0]',
                'b.C[1]',
                'b.C[1][0]',
                'B',
                'B[0]',
                'B[0].v',
                'B[0].V',
                'B[0].V[0]',
                'B[0].V[1]',
                'B[0].V[1][0]',
                'B[0].c',
                'B[0].C',
                'B[0].C[0]',
                'B[0].C[1]',
                'B[0].C[1][0]',
                'B[1]',
                'B[1][0]',
                'B[1][0].v',
                'B[1][0].V',
                'B[1][0].V[0]',
                'B[1][0].V[1]',
                'B[1][0].V[1][0]',
                'B[1][0].c',
                'B[1][0].C',
                'B[1][0].C[0]',
                'B[1][0].C[1]',
                'B[1][0].C[1][0]',
                'j',
                'J',
                'J[0]',
                'J[1]',
                'J[1][0]',
                'J[1][0].b',
                'J[1][0].b.v',
                'k',
                'K',
                'K[0]',
                'K[0].v',
                'K[0].V',
                'K[0].V[0]',
                'K[0].V[1]',
                'K[0].V[1][0]',
                'K[0].c',
                'K[0].C',
                'K[0].C[0]',
                'K[0].C[1]',
                'K[0].C[1][0]',
                'K[0].b',
                'K[0].b.v',
                'K[0].b.V',
                'K[0].b.V[0]',
                'K[0].b.V[1]',
                'K[0].b.V[1][0]',
                'K[0].b.c',
                'K[0].b.C',
                'K[0].b.C[0]',
                'K[0].b.C[1]',
                'K[0].b.C[1][0]',
                'K[0].B',
                'K[0].B[0]',
                'K[0].B[0].v',
                'K[0].B[0].V',
                'K[0].B[0].V[0]',
                'K[0].B[0].V[1]',
                'K[0].B[0].V[1][0]',
                'K[0].B[0].c',
                'K[0].B[0].C',
                'K[0].B[0].C[0]',
                'K[0].B[0].C[1]',
                'K[0].B[0].C[1][0]',
                'K[0].B[1]',
                'K[0].B[1][0]',
                'K[0].B[1][0].v',
                'K[0].B[1][0].V',
                'K[0].B[1][0].V[0]',
                'K[0].B[1][0].V[1]',
                'K[0].B[1][0].V[1][0]',
                'K[0].B[1][0].c',
                'K[0].B[1][0].C',
                'K[0].B[1][0].C[0]',
                'K[0].B[1][0].C[1]',
                'K[0].B[1][0].C[1][0]',
                'K[0].j',
                'K[0].J',
                'K[0].J[0]',
                'K[0].J[1]',
                'K[0].J[1][0]',
                'K[0].J[1][0].b',
                'K[0].J[1][0].b.v',
            ],
        )

        order = list(
            str(obj)
            for obj in pmo.preorder_traversal(
                model, descend=lambda x: (x is not model.k) and (x is not model.K)
            )
        )
        self.assertEqual(
            order,
            [
                '<block>',
                'v',
                'V',
                'V[0]',
                'V[1]',
                'V[1][0]',
                'c',
                'C',
                'C[0]',
                'C[1]',
                'C[1][0]',
                'b',
                'b.v',
                'b.V',
                'b.V[0]',
                'b.V[1]',
                'b.V[1][0]',
                'b.c',
                'b.C',
                'b.C[0]',
                'b.C[1]',
                'b.C[1][0]',
                'B',
                'B[0]',
                'B[0].v',
                'B[0].V',
                'B[0].V[0]',
                'B[0].V[1]',
                'B[0].V[1][0]',
                'B[0].c',
                'B[0].C',
                'B[0].C[0]',
                'B[0].C[1]',
                'B[0].C[1][0]',
                'B[1]',
                'B[1][0]',
                'B[1][0].v',
                'B[1][0].V',
                'B[1][0].V[0]',
                'B[1][0].V[1]',
                'B[1][0].V[1][0]',
                'B[1][0].c',
                'B[1][0].C',
                'B[1][0].C[0]',
                'B[1][0].C[1]',
                'B[1][0].C[1][0]',
                'j',
                'J',
                'J[0]',
                'J[1]',
                'J[1][0]',
                'J[1][0].b',
                'J[1][0].b.v',
                'k',
                'K',
            ],
        )

        order = list(str(obj) for obj in pmo.preorder_traversal(model, ctype=IBlock))
        self.assertEqual(
            order,
            [
                '<block>',
                'b',
                'B',
                'B[0]',
                'B[1]',
                'B[1][0]',
                'j',
                'J',
                'J[0]',
                'J[1]',
                'J[1][0]',
                'J[1][0].b',
                'k',
                'K',
                'K[0]',
                'K[0].b',
                'K[0].B',
                'K[0].B[0]',
                'K[0].B[1]',
                'K[0].B[1][0]',
                'K[0].j',
                'K[0].J',
                'K[0].J[0]',
                'K[0].J[1]',
                'K[0].J[1][0]',
                'K[0].J[1][0].b',
            ],
        )

        order = list(str(obj) for obj in pmo.preorder_traversal(model, ctype=IVariable))
        self.assertEqual(
            order,
            [
                '<block>',
                'v',
                'V',
                'V[0]',
                'V[1]',
                'V[1][0]',
                'b',
                'b.v',
                'b.V',
                'b.V[0]',
                'b.V[1]',
                'b.V[1][0]',
                'B',
                'B[0]',
                'B[0].v',
                'B[0].V',
                'B[0].V[0]',
                'B[0].V[1]',
                'B[0].V[1][0]',
                'B[1]',
                'B[1][0]',
                'B[1][0].v',
                'B[1][0].V',
                'B[1][0].V[0]',
                'B[1][0].V[1]',
                'B[1][0].V[1][0]',
                'j',
                'J',
                'J[0]',
                'J[1]',
                'J[1][0]',
                'J[1][0].b',
                'J[1][0].b.v',
                'k',
                'K',
                'K[0]',
                'K[0].v',
                'K[0].V',
                'K[0].V[0]',
                'K[0].V[1]',
                'K[0].V[1][0]',
                'K[0].b',
                'K[0].b.v',
                'K[0].b.V',
                'K[0].b.V[0]',
                'K[0].b.V[1]',
                'K[0].b.V[1][0]',
                'K[0].B',
                'K[0].B[0]',
                'K[0].B[0].v',
                'K[0].B[0].V',
                'K[0].B[0].V[0]',
                'K[0].B[0].V[1]',
                'K[0].B[0].V[1][0]',
                'K[0].B[1]',
                'K[0].B[1][0]',
                'K[0].B[1][0].v',
                'K[0].B[1][0].V',
                'K[0].B[1][0].V[0]',
                'K[0].B[1][0].V[1]',
                'K[0].B[1][0].V[1][0]',
                'K[0].j',
                'K[0].J',
                'K[0].J[0]',
                'K[0].J[1]',
                'K[0].J[1][0]',
                'K[0].J[1][0].b',
                'K[0].J[1][0].b.v',
            ],
        )

    def test_components(self):
        model = self.model.clone()
        checked = []

        def descend_into(x):
            self.assertTrue(x._is_heterogeneous_container)
            checked.append(x.name)
            return True

        order = list(str(obj) for obj in model.components(descend_into=descend_into))
        self.assertEqual(
            checked,
            [
                'b',
                'B[0]',
                'B[1][0]',
                'j',
                'J[0]',
                'J[1][0]',
                'J[1][0].b',
                'k',
                'K[0]',
                'K[0].b',
                'K[0].B[0]',
                'K[0].B[1][0]',
                'K[0].j',
                'K[0].J[0]',
                'K[0].J[1][0]',
                'K[0].J[1][0].b',
            ],
        )
        self.assertEqual(
            order,
            [
                'v',
                'V[0]',
                'V[1][0]',
                'c',
                'C[0]',
                'C[1][0]',
                'b',
                'b.v',
                'b.V[0]',
                'b.V[1][0]',
                'b.c',
                'b.C[0]',
                'b.C[1][0]',
                'B[0]',
                'B[0].v',
                'B[0].V[0]',
                'B[0].V[1][0]',
                'B[0].c',
                'B[0].C[0]',
                'B[0].C[1][0]',
                'B[1][0]',
                'B[1][0].v',
                'B[1][0].V[0]',
                'B[1][0].V[1][0]',
                'B[1][0].c',
                'B[1][0].C[0]',
                'B[1][0].C[1][0]',
                'j',
                'J[0]',
                'J[1][0]',
                'J[1][0].b',
                'J[1][0].b.v',
                'k',
                'K[0]',
                'K[0].v',
                'K[0].V[0]',
                'K[0].V[1][0]',
                'K[0].c',
                'K[0].C[0]',
                'K[0].C[1][0]',
                'K[0].b',
                'K[0].b.v',
                'K[0].b.V[0]',
                'K[0].b.V[1][0]',
                'K[0].b.c',
                'K[0].b.C[0]',
                'K[0].b.C[1][0]',
                'K[0].B[0]',
                'K[0].B[0].v',
                'K[0].B[0].V[0]',
                'K[0].B[0].V[1][0]',
                'K[0].B[0].c',
                'K[0].B[0].C[0]',
                'K[0].B[0].C[1][0]',
                'K[0].B[1][0]',
                'K[0].B[1][0].v',
                'K[0].B[1][0].V[0]',
                'K[0].B[1][0].V[1][0]',
                'K[0].B[1][0].c',
                'K[0].B[1][0].C[0]',
                'K[0].B[1][0].C[1][0]',
                'K[0].j',
                'K[0].J[0]',
                'K[0].J[1][0]',
                'K[0].J[1][0].b',
                'K[0].J[1][0].b.v',
            ],
        )
        vlist = [str(obj) for obj in model.components(ctype=IVariable)]
        self.assertEqual(len(vlist), len(set(vlist)))
        clist = [str(obj) for obj in model.components(ctype=IConstraint)]
        self.assertEqual(len(clist), len(set(clist)))
        blist = [str(obj) for obj in model.components(ctype=IBlock)]
        self.assertEqual(len(blist), len(set(blist)))
        jlist = [str(obj) for obj in model.components(ctype=IJunk)]
        self.assertEqual(len(jlist), len(set(jlist)))

        for l1, l2 in itertools.product([vlist, clist, blist, jlist], repeat=2):
            if l1 is l2:
                continue
            self.assertEqual(set(l1).intersection(set(l2)), set([]))
        self.assertEqual(len(vlist) + len(clist) + len(blist) + len(jlist), len(order))

    def test_getname(self):
        model = self.model.clone()
        self.assertEqual(model.J[1][0].b.v.getname(fully_qualified=True), 'J[1][0].b.v')
        self.assertEqual(
            model.J[1][0].b.v.getname(fully_qualified=True, relative_to=model.J[1][0]),
            'b.v',
        )
        self.assertEqual(
            model.J[1][0].b.v.getname(fully_qualified=True, relative_to=model.J[1]),
            '[0].b.v',
        )
        self.assertEqual(
            model.J[1][0].b.v.getname(fully_qualified=True, relative_to=model.J),
            '[1][0].b.v',
        )
        self.assertEqual(
            model.J[1][0].b.v.getname(fully_qualified=True, relative_to=model),
            'J[1][0].b.v',
        )

    def test_heterogeneous_containers(self):
        order = list(str(obj) for obj in heterogeneous_containers(self.model.V))
        self.assertEqual(order, [])
        order = list(str(obj) for obj in heterogeneous_containers(self.model.v))
        self.assertEqual(order, [])

        order = list(str(obj) for obj in heterogeneous_containers(self.model))
        self.assertEqual(
            order,
            [
                '<block>',
                'b',
                'B[0]',
                'B[1][0]',
                'k',
                'K[0]',
                'K[0].b',
                'K[0].B[0]',
                'K[0].B[1][0]',
                'K[0].j',
                'K[0].J[0]',
                'K[0].J[1][0]',
                'K[0].J[1][0].b',
                'j',
                'J[0]',
                'J[1][0]',
                'J[1][0].b',
            ],
        )

        def f(x):
            # do not descend below heterogeneous containers
            # stored on self.model
            self.assertTrue(x._is_heterogeneous_container)
            parent = x.parent
            while parent is not None:
                if parent is self.model:
                    return False
                parent = parent.parent
            return True

        order1 = list(
            str(obj) for obj in heterogeneous_containers(self.model, descend_into=f)
        )
        order2 = list(
            str(obj)
            for obj in heterogeneous_containers(
                self.model, descend_into=lambda x: True if (x is self.model) else False
            )
        )
        self.assertEqual(order1, order2)
        self.assertEqual(
            order1,
            ['<block>', 'b', 'B[0]', 'B[1][0]', 'k', 'K[0]', 'j', 'J[0]', 'J[1][0]'],
        )
        order = list(
            str(obj) for obj in heterogeneous_containers(self.model, ctype=IBlock)
        )
        self.assertEqual(
            order,
            [
                '<block>',
                'b',
                'B[0]',
                'B[1][0]',
                'k',
                'K[0]',
                'K[0].b',
                'K[0].B[0]',
                'K[0].B[1][0]',
                'K[0].J[1][0].b',
                'J[1][0].b',
            ],
        )
        order = list(
            str(obj) for obj in heterogeneous_containers(self.model, ctype=IJunk)
        )
        self.assertEqual(
            order, ['K[0].j', 'K[0].J[0]', 'K[0].J[1][0]', 'j', 'J[0]', 'J[1][0]']
        )
        order = list(
            str(obj) for obj in heterogeneous_containers(self.model.K, ctype=IJunk)
        )
        self.assertEqual(order, ['K[0].j', 'K[0].J[0]', 'K[0].J[1][0]'])
        order = list(
            str(obj) for obj in heterogeneous_containers(self.model.K[0], ctype=IJunk)
        )
        self.assertEqual(order, ['K[0].j', 'K[0].J[0]', 'K[0].J[1][0]'])
        order = list(
            str(obj) for obj in heterogeneous_containers(self.model.K[0].j, ctype=IJunk)
        )
        self.assertEqual(order, ['K[0].j'])
        order = list(
            str(obj)
            for obj in heterogeneous_containers(self.model.K[0].j, ctype=IBlock)
        )
        self.assertEqual(order, [])


class TestMisc(unittest.TestCase):
    def test_reserved_attributes(self):
        b = block()
        self.assertTrue(len(block._block_reserved_words) > 0)
        for name in block._block_reserved_words:
            with self.assertRaises(ValueError):
                setattr(b, name, 1)
        with self.assertRaises(ValueError):
            b._active = 1
        with self.assertRaises(ValueError):
            b._parent = 1
        with self.assertRaises(ValueError):
            b._storage_key = 1
        with self.assertRaises(AttributeError):
            b.active = 1

    def test_pprint(self):
        # Not really testing what the output is, just that
        # an error does not occur. The pprint functionality
        # is still in the early stages.
        B = block()
        B.s = suffix()
        B.b = block()
        B.v = variable()
        pmo.pprint(B)
        B.c = constraint()
        B.e = expression()
        B.o = objective()
        B.p = parameter()
        B.s = sos([])
        pmo.pprint(B)
        b = block()
        b.B = B
        pmo.pprint(B)
        pmo.pprint(b)
        m = block()
        m.b = b
        pmo.pprint(B)
        pmo.pprint(b)
        pmo.pprint(m)

    def test_ctype(self):
        b = block()
        self.assertIs(b.ctype, IBlock)
        self.assertIs(type(b), block)
        self.assertIs(type(b)._ctype, IBlock)

    def test_write(self):
        b = block()
        b.v = variable()
        b.y = variable(value=1.0, fixed=True)
        b.p = parameter(value=2.0)
        b.e = expression(b.v * b.p + 1)
        b.eu = data_expression(b.p**2 + 10)
        b.el = data_expression(-b.p**2 - 10)
        b.o = objective(b.v + b.y)
        b.c1 = constraint((b.el + 1, b.e + 2, b.eu + 2))
        b.c2 = constraint(lb=b.el, body=b.v)
        b.c3 = constraint(body=b.v, ub=b.eu)

        fid, fname = tempfile.mkstemp(suffix='.lp')
        os.close(fid)
        assert fname.endswith('.lp')
        smap = b.write(fname)
        os.remove(fname)
        self.assertTrue(isinstance(smap, SymbolMap))

        fid, fname = tempfile.mkstemp(suffix='.lp')
        os.close(fid)
        assert fname.endswith('.lp')
        smap = b.write(fname, format='nl')
        os.remove(fname)
        self.assertTrue(isinstance(smap, SymbolMap))

        fid, fname = tempfile.mkstemp(suffix='.sdfsdfsf')
        os.close(fid)
        assert fname.endswith('.sdfsdfsf')
        with self.assertRaises(ValueError):
            b.write(fname)
        os.remove(fname)

        fid, fname = tempfile.mkstemp(suffix='.sdfsdfsf')
        os.close(fid)
        assert fname.endswith('.sdfsdfsf')
        with self.assertRaises(ValueError):
            b.write(fname, format="sdfdsfsdfsf")
        os.remove(fname)

    def test_load_solution(self):
        sm = SymbolMap()
        m = block()
        sm.addSymbol(m, 'm')
        m.v = variable()
        sm.addSymbol(m.v, 'v')
        m.c = constraint()
        sm.addSymbol(m.c, 'c')
        m.o = objective()
        sm.addSymbol(m.o, 'o')
        m.vsuffix = suffix(direction=suffix.IMPORT)
        m.osuffix = suffix(direction=suffix.IMPORT)
        m.csuffix = suffix(direction=suffix.IMPORT)
        m.msuffix = suffix(direction=suffix.IMPORT)

        soln = Solution()
        soln.symbol_map = sm
        soln.variable['v'] = {"Value": 1.0, "vsuffix": 'v'}
        soln.variable['ONE_VAR_CONSTANT'] = None
        soln.constraint['c'] = {"csuffix": 'c'}
        soln.constraint['ONE_VAR_CONSTANT'] = None
        soln.objective['o'] = {"osuffix": 'o'}
        soln.problem["msuffix"] = 'm'

        m.load_solution(soln)

        self.assertEqual(m.v.value, 1.0)
        self.assertEqual(m.csuffix[m.c], 'c')
        self.assertEqual(m.osuffix[m.o], 'o')
        self.assertEqual(m.msuffix[m], 'm')

        soln.variable['vv'] = {"Value": 1.0, "vsuffix": 'v'}
        with self.assertRaises(KeyError):
            m.load_solution(soln)
        del soln.variable['vv']

        soln.constraint['cc'] = {"csuffix": 'c'}
        with self.assertRaises(KeyError):
            m.load_solution(soln)
        del soln.constraint['cc']

        soln.objective['oo'] = {"osuffix": 'o'}
        with self.assertRaises(KeyError):
            m.load_solution(soln)
        del soln.objective['oo']

        m.v.fix()
        with self.assertRaises(ValueError):
            m.load_solution(soln, allow_consistent_values_for_fixed_vars=False)

        m.v.fix(1.1)
        m.load_solution(
            soln,
            allow_consistent_values_for_fixed_vars=True,
            comparison_tolerance_for_fixed_vars=0.5,
        )
        m.v.fix(1.1)
        with self.assertRaises(ValueError):
            m.load_solution(
                soln,
                allow_consistent_values_for_fixed_vars=True,
                comparison_tolerance_for_fixed_vars=0.05,
            )

        del soln.variable['v']

        m.v.free()
        m.v.value = None
        m.load_solution(soln)
        self.assertEqual(m.v.stale, True)
        self.assertEqual(m.v.value, None)

        soln.default_variable_value = 1.0
        m.load_solution(soln)
        self.assertEqual(m.v.stale, False)
        self.assertEqual(m.v.value, 1.0)

        m.v.fix(1.0)
        with self.assertRaises(ValueError):
            m.load_solution(soln, allow_consistent_values_for_fixed_vars=False)

        m.v.fix(1.1)
        m.load_solution(
            soln,
            allow_consistent_values_for_fixed_vars=True,
            comparison_tolerance_for_fixed_vars=0.5,
        )
        m.v.fix(1.1)
        with self.assertRaises(ValueError):
            m.load_solution(
                soln,
                allow_consistent_values_for_fixed_vars=True,
                comparison_tolerance_for_fixed_vars=0.05,
            )

    # a temporary test to make sure solve and load
    # functionality work (will be moved elsewhere in the
    # future)
    def test_solve_load(self):
        b = block()
        b.v = variable()
        b.y = variable(value=1.0, fixed=True)
        b.p = parameter(value=2.0)
        b.e = expression(b.v * b.p + 1)
        b.eu = data_expression(b.p**2 + 10)
        b.el = data_expression(-b.p**2 - 10)
        b.o = objective(b.v + b.y)
        b.c1 = constraint((b.el + 1, b.e + 2, b.eu + 2))
        b.c2 = constraint(lb=b.el, body=b.v)
        b.c3 = constraint(body=b.v, ub=b.eu)
        b.dual = suffix(direction=suffix.IMPORT)

        import pyomo.environ
        from pyomo.opt.base.solvers import UnknownSolver
        from pyomo.opt import SolverFactory
        from pyomo.opt import SolverStatus, TerminationCondition

        opt = SolverFactory("glpk")
        if isinstance(opt, UnknownSolver) or (not opt.available()):
            raise unittest.SkipTest("glpk solver not available")
        status = opt.solve(b)
        self.assertEqual(status.solver.status, SolverStatus.ok)
        self.assertEqual(
            status.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(b.o(), -7, places=5)
        self.assertAlmostEqual(b.v(), -8, places=5)
        self.assertAlmostEqual(b.y(), 1.0, places=5)

        opt = SolverFactory("glpk")
        if isinstance(opt, UnknownSolver) or (not opt.available()):
            raise unittest.SkipTest("glpk solver not available")
        status = opt.solve(b, symbolic_solver_labels=True)
        self.assertEqual(status.solver.status, SolverStatus.ok)
        self.assertEqual(
            status.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(b.o(), -7, places=5)
        self.assertAlmostEqual(b.v(), -8, places=5)
        self.assertAlmostEqual(b.y(), 1.0, places=5)

        opt = SolverFactory("ipopt")
        if isinstance(opt, UnknownSolver) or (not opt.available()):
            raise unittest.SkipTest("ipopt solver not available")
        status = opt.solve(b)
        self.assertEqual(status.solver.status, SolverStatus.ok)
        self.assertEqual(
            status.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(b.o(), -7, places=5)
        self.assertAlmostEqual(b.v(), -8, places=5)
        self.assertAlmostEqual(b.y(), 1.0, places=5)

        opt = SolverFactory("ipopt")
        if isinstance(opt, UnknownSolver):
            raise unittest.SkipTest("ipopt solver not available")
        status = opt.solve(b, symbolic_solver_labels=True)
        self.assertEqual(status.solver.status, SolverStatus.ok)
        self.assertEqual(
            status.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertAlmostEqual(b.o(), -7, places=5)
        self.assertAlmostEqual(b.v(), -8, places=5)
        self.assertAlmostEqual(b.y(), 1.0, places=5)

    def test_traversal(self):
        b = block()
        b.v = variable()
        b.c1 = constraint()
        b.c1.deactivate()
        b.c2 = constraint_list()
        b.c2.append(constraint_list())
        b.B = block_list()
        b.B.append(block_list())
        b.B[0].append(block())
        b.B[0][0].c = constraint()
        b.B[0][0].b = block()
        b.B[0].deactivate()
        b._activate_large_storage_mode()

        def descend(obj):
            self.assertTrue(obj._is_container)
            return True

        self.assertEqual(
            [
                obj.name
                for obj in pmo.preorder_traversal(b, active=None, descend=descend)
            ],
            [
                None,
                'v',
                'c1',
                'c2',
                'c2[0]',
                'B',
                'B[0]',
                'B[0][0]',
                'B[0][0].c',
                'B[0][0].b',
            ],
        )
        self.assertEqual(
            [obj.name for obj in pmo.preorder_traversal(b, descend=descend)],
            [None, 'v', 'c2', 'c2[0]', 'B'],
        )
        self.assertEqual(
            [
                obj.name
                for obj in pmo.preorder_traversal(b, active=True, descend=descend)
            ],
            [None, 'v', 'c2', 'c2[0]', 'B'],
        )

        self.assertEqual(
            [
                obj.name
                for obj in pmo.preorder_traversal(
                    b, active=None, ctype=IConstraint, descend=descend
                )
            ],
            [
                None,
                'c1',
                'c2',
                'c2[0]',
                'B',
                'B[0]',
                'B[0][0]',
                'B[0][0].c',
                'B[0][0].b',
            ],
        )
        self.assertEqual(
            [
                obj.name
                for obj in pmo.preorder_traversal(b, ctype=IConstraint, descend=descend)
            ],
            [None, 'c2', 'c2[0]', 'B'],
        )
        self.assertEqual(
            [
                obj.name
                for obj in pmo.preorder_traversal(
                    b, ctype=IConstraint, active=True, descend=descend
                )
            ],
            [None, 'c2', 'c2[0]', 'B'],
        )

        m = pmo.block()
        m.B = pmo.block_list()
        m.B.append(pmo.block())
        m.B[0].v = pmo.variable()

        self.assertEqual(
            [obj.name for obj in pmo.preorder_traversal(m, ctype=IVariable)],
            [None, 'B', 'B[0]', 'B[0].v'],
        )
        self.assertEqual(
            [obj.name for obj in pmo.preorder_traversal(m.B, ctype=IVariable)],
            ['B', 'B[0]', 'B[0].v'],
        )
        self.assertEqual(
            [obj.name for obj in pmo.preorder_traversal(m.B[0], ctype=IVariable)],
            ['B[0]', 'B[0].v'],
        )
        self.assertEqual(
            [obj.name for obj in pmo.preorder_traversal(m.B[0].v, ctype=IVariable)],
            ['B[0].v'],
        )

    # test how clone behaves when there are
    # references to components on a different block
    def test_clone1(self):
        b = block()
        b.v = variable()
        b.b = block()
        b.b.e = expression(b.v**2)
        b.b.v = variable()
        b.bdict = block_dict()
        b.bdict[0] = block()
        b.bdict[0].e = expression(b.v**2)
        b.blist = block_list()
        b.blist.append(block())
        b.blist[0].e = expression(b.v**2 + b.b.v**2)

        bc = b.clone()
        self.assertIsNot(b.v, bc.v)
        self.assertIsNot(b.b.e, bc.b.e)
        self.assertIsNot(b.bdict[0].e, bc.bdict[0].e)
        self.assertIsNot(b.blist[0].e, bc.blist[0].e)

        #
        # check that the expressions on cloned sub-blocks
        # reference the original variables for blocks "out-of-scope"
        #
        b_b = b.b.clone()
        self.assertIsNot(b_b.e, b.b.e)
        self.assertTrue(len(_collect_expr_components(b.b.e.expr)) == 1)
        self.assertIs(list(_collect_expr_components(b.b.e.expr).values())[0], b.v)
        self.assertTrue(len(_collect_expr_components(b_b.e.expr)) == 1)
        self.assertIs(list(_collect_expr_components(b_b.e.expr).values())[0], b.v)

        b_bdict0 = b.bdict[0].clone()
        self.assertIsNot(b_bdict0.e, b.bdict[0].e)
        self.assertTrue(len(_collect_expr_components(b.bdict[0].e.expr)) == 1)
        self.assertIs(
            list(_collect_expr_components(b.bdict[0].e.expr).values())[0], b.v
        )
        self.assertTrue(len(_collect_expr_components(b_bdict0.e.expr)) == 1)
        self.assertIs(list(_collect_expr_components(b_bdict0.e.expr).values())[0], b.v)

        b_blist0 = b.blist[0].clone()
        self.assertIsNot(b_blist0.e, b.blist[0].e)
        self.assertTrue(len(_collect_expr_components(b.blist[0].e.expr)) == 2)
        self.assertEqual(
            sorted(
                list(
                    id(v_)
                    for v_ in _collect_expr_components(b.blist[0].e.expr).values()
                )
            ),
            sorted(list(id(v_) for v_ in [b.v, b.b.v])),
        )
        self.assertTrue(len(_collect_expr_components(b_blist0.e.expr)) == 2)
        self.assertEqual(
            sorted(
                list(
                    id(v_) for v_ in _collect_expr_components(b_blist0.e.expr).values()
                )
            ),
            sorted(list(id(v_) for v_ in [b.v, b.b.v])),
        )

    # test bulk clone behavior
    def test_clone2(self):
        b = block()
        b.v = variable()
        b.vdict = variable_dict(((i, variable()) for i in range(10)))
        b.vlist = variable_list(variable() for i in range(10))
        b.o = objective(b.v + b.vdict[0] + b.vlist[0])
        b.odict = objective_dict(((i, objective(b.v + b.vdict[i])) for i in b.vdict))
        b.olist = objective_list(objective(b.v + v_) for i, v_ in enumerate(b.vdict))
        b.c = constraint(b.v >= 1)
        b.cdict = constraint_dict(((i, constraint(b.vdict[i] == i)) for i in b.vdict))
        b.clist = constraint_list(
            constraint((0, v_, i)) for i, v_ in enumerate(b.vlist)
        )
        b.p = parameter()
        b.pdict = parameter_dict(((i, parameter(i)) for i in b.vdict))
        b.plist = parameter_list(parameter(i) for i in range(len(b.vlist)))
        b.e = expression(b.v * b.p + 1)
        b.edict = expression_dict(
            ((i, expression(b.vdict[i] * b.pdict[i] + 1)) for i in b.vdict)
        )
        b.elist = expression_list(
            expression(v_ * b.plist[i] + 1) for i, v_ in enumerate(b.vlist)
        )

        self.assertIs(b.parent, None)

        #
        # clone the block
        #
        bc = b.clone()
        self.assertIs(bc.parent, None)
        self.assertIsNot(b, bc)
        self.assertTrue(len(list(b.children())) > 0)
        self.assertEqual(len(list(b.children())), len(list(bc.children())))
        for c1, c2 in zip(b.children(), bc.children()):
            self.assertIs(c1.parent, b)
            self.assertIs(c2.parent, bc)
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)

        self.assertEqual(len(list(b.components())), len(list(bc.components())))
        for c1, c2 in zip(b.components(), bc.components()):
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)
            if hasattr(c1, 'expr'):
                self.assertIsNot(c1.expr, c2.expr)
                self.assertEqual(str(c1.expr), str(c2.expr))
                self.assertEqual(
                    len(_collect_expr_components(c1.expr)),
                    len(_collect_expr_components(c2.expr)),
                )
                for subc1, subc2 in zip(
                    _collect_expr_components(c1.expr).values(),
                    _collect_expr_components(c2.expr).values(),
                ):
                    self.assertIsNot(subc1, subc2)
                    self.assertEqual(subc1.name, subc2.name)

        bc_init = bc.clone()
        b.bc = bc
        self.assertIs(b.parent, None)
        self.assertIs(bc.parent, b)
        #
        # clone the block with the newly added sub-block
        #

        bcc = b.clone()
        self.assertIsNot(b, bcc)
        self.assertEqual(len(list(b.children())), len(list(bcc.children())))
        for c1, c2 in zip(b.children(), bcc.children()):
            self.assertIs(c1.parent, b)
            self.assertIs(c2.parent, bcc)
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)

        self.assertEqual(len(list(b.components())), len(list(bcc.components())))
        self.assertTrue(hasattr(bcc, 'bc'))
        for c1, c2 in zip(b.components(), bcc.components()):
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)
            if hasattr(c1, 'expr'):
                self.assertIsNot(c1.expr, c2.expr)
                self.assertEqual(str(c1.expr), str(c2.expr))
                self.assertEqual(
                    len(_collect_expr_components(c1.expr)),
                    len(_collect_expr_components(c2.expr)),
                )
                for subc1, subc2 in zip(
                    _collect_expr_components(c1.expr).values(),
                    _collect_expr_components(c2.expr).values(),
                ):
                    self.assertIsNot(subc1, subc2)
                    self.assertEqual(subc1.name, subc2.name)

        #
        # clone the sub-block
        #
        sub_bc = b.bc.clone()
        self.assertIs(sub_bc.parent, None)
        self.assertIs(bc_init.parent, None)
        self.assertIs(bc.parent, b)
        self.assertIs(b.parent, None)

        self.assertIsNot(bc_init, sub_bc)
        self.assertIsNot(bc, sub_bc)
        self.assertEqual(len(list(bc_init.children())), len(list(sub_bc.children())))
        for c1, c2 in zip(bc_init.children(), sub_bc.children()):
            self.assertIs(c1.parent, bc_init)
            self.assertIs(c2.parent, sub_bc)
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)

        self.assertEqual(
            len(list(bc_init.components())), len(list(sub_bc.components()))
        )
        for c1, c2 in zip(bc_init.components(), sub_bc.components()):
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)
            if hasattr(c1, 'expr'):
                self.assertIsNot(c1.expr, c2.expr)
                self.assertEqual(str(c1.expr), str(c2.expr))
                self.assertEqual(
                    len(_collect_expr_components(c1.expr)),
                    len(_collect_expr_components(c2.expr)),
                )
                for subc1, subc2 in zip(
                    _collect_expr_components(c1.expr).values(),
                    _collect_expr_components(c2.expr).values(),
                ):
                    self.assertIsNot(subc1, subc2)
                    self.assertEqual(subc1.name, subc2.name)

    def test_activate(self):
        b = block()
        self.assertEqual(b.active, True)
        b.deactivate()
        self.assertEqual(b.active, False)
        c = constraint()
        v = variable()
        self.assertEqual(c.active, True)
        c.deactivate()
        self.assertEqual(c.active, False)
        b.c = c
        b.v = v
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)
        del b.c
        c.activate()
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, False)
        b.c = c
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, False)

        bdict = block_dict()
        self.assertEqual(bdict.active, True)
        bdict.deactivate()
        self.assertEqual(bdict.active, False)
        bdict[None] = b
        self.assertEqual(bdict.active, False)
        del bdict[None]
        self.assertEqual(bdict.active, False)
        b.activate()
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, True)
        bdict[None] = b
        self.assertEqual(bdict.active, False)

        bdict.deactivate()
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(bdict.active, False)

        bdict.deactivate(shallow=False)
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(bdict.active, False)

        b.activate()
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, True)
        self.assertEqual(bdict.active, False)

        b.activate(shallow=False)
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(bdict.active, False)

        bdict.deactivate()
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(bdict.active, False)

        bdict.deactivate(shallow=False)
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(bdict.active, False)

        bdict.activate()
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(bdict.active, True)

        bdict.activate(shallow=False)
        self.assertEqual(c.active, True)
        self.assertEqual(b.active, True)
        self.assertEqual(bdict.active, True)

        bdict.deactivate(shallow=False)
        self.assertEqual(c.active, False)
        self.assertEqual(b.active, False)
        self.assertEqual(bdict.active, False)

    # this is a randomized test
    def test_ordering(self):
        b = block()
        attr_types = [
            variable,
            constraint,
            parameter,
            expression,
            data_expression,
            objective,
            variable,
            block,
        ]
        key_types = [float, int]
        keys = collections.deque()
        objs = collections.deque()
        for i in range(100):
            obj = random.choice(attr_types)()
            key = str(random.choice(key_types)(i))
            setattr(b, key, obj)
            keys.append(key)
            objs.append(obj)
            self.assertEqual(len(list(b.children())), len(objs))

        for i in range(100):
            children = list(b.children())
            for key, obj, child in zip(keys, objs, children):
                assert obj is child
                assert getattr(b, key) is child
                assert key == child.storage_key
            setattr(b, children[0].storage_key, children[0])
            keys.rotate(-1)
            objs.rotate(-1)


class _Test_block_base:
    _children = None
    _child_key = None
    _components_no_descend = None
    _components = None
    _blocks_no_descend = None
    _blocks = None
    _block = None

    def test_overwrite_warning(self):
        b = self._block.clone()
        name = "x"
        while hasattr(b, name):
            name += "x"
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.core'):
            setattr(b, name, variable())
            setattr(b, name, getattr(b, name))
        assert out.getvalue() == "", str(out.getvalue())
        with LoggingIntercept(out, 'pyomo.core'):
            setattr(b, name, variable())
        assert out.getvalue() == (
            "Implicitly replacing attribute %s (type=variable) "
            "on block with new object (type=variable). This "
            "is usually indicative of a modeling error. "
            "To avoid this warning, delete the original "
            "object from the block before assigning a new "
            "object.\n" % (name)
        )
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.core'):
            setattr(b, name, 1.0)
        assert out.getvalue() == (
            "Implicitly replacing attribute %s (type=variable) "
            "on block with new object (type=float). This "
            "is usually indicative of a modeling error. "
            "To avoid this warning, delete the original "
            "object from the block before assigning a new "
            "object.\n" % (name)
        )

    def test_clone(self):
        b = self._block
        bc = b.clone()
        self.assertIsNot(b, bc)
        self.assertEqual(len(list(b.children())), len(list(bc.children())))
        for c1, c2 in zip(b.children(), bc.children()):
            self.assertIs(c1.parent, b)
            self.assertIs(c2.parent, bc)
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)

        self.assertEqual(len(list(b.components())), len(list(bc.components())))
        for c1, c2 in zip(b.components(), bc.components()):
            self.assertIsNot(c1, c2)
            self.assertEqual(c1.name, c2.name)

    def test_pickle(self):
        b = pickle.loads(pickle.dumps(self._block))
        self.assertEqual(
            len(list(pmo.preorder_traversal(b, active=None))), len(self._names) + 1
        )
        names = pmo.generate_names(b, active=None)
        self.assertEqual(sorted(names.values()), sorted(self._names.values()))

    def test_preorder_traversal(self):
        # this first test makes failures a
        # little easier to debug
        self.assertEqual(
            [str(obj) for obj in pmo.preorder_traversal(self._block, active=None)],
            [str(obj) for obj in self._preorder],
        )
        self.assertEqual(
            [id(obj) for obj in pmo.preorder_traversal(self._block, active=None)],
            [id(obj) for obj in self._preorder],
        )

        # this first test makes failures a
        # little easier to debug
        self.assertEqual(
            [
                str(obj)
                for obj in pmo.preorder_traversal(
                    self._block, active=None, ctype=IVariable
                )
            ],
            [str(obj) for obj in self._preorder if obj.ctype in (IBlock, IVariable)],
        )
        self.assertEqual(
            [
                id(obj)
                for obj in pmo.preorder_traversal(
                    self._block, active=None, ctype=IVariable
                )
            ],
            [id(obj) for obj in self._preorder if obj.ctype in (IBlock, IVariable)],
        )

    def test_preorder_traversal_descend_check(self):
        def descend(x):
            self.assertTrue(x._is_container)
            return True

        order = list(pmo.preorder_traversal(self._block, active=None, descend=descend))
        # this first test makes failures a
        # little easier to debug
        self.assertEqual(
            [str(obj) for obj in order], [str(obj) for obj in self._preorder]
        )
        self.assertEqual(
            [id(obj) for obj in order], [id(obj) for obj in self._preorder]
        )

        def descend(x):
            self.assertTrue(x._is_container)
            return True

        order = list(
            pmo.preorder_traversal(
                self._block, active=None, ctype=IVariable, descend=descend
            )
        )
        # this first test makes failures a
        # little easier to debug
        self.assertEqual(
            [str(obj) for obj in order],
            [str(obj) for obj in self._preorder if obj.ctype in (IBlock, IVariable)],
        )
        self.assertEqual(
            [id(obj) for obj in order],
            [id(obj) for obj in self._preorder if obj.ctype in (IBlock, IVariable)],
        )

        def descend(x):
            if x.parent is self._block:
                return False
            return True

        order = list(pmo.preorder_traversal(self._block, active=None, descend=descend))
        # this first test makes failures a
        # little easier to debug
        self.assertEqual(
            [str(obj) for obj in order],
            [
                str(obj)
                for obj in self._preorder
                if (obj.parent is None) or (obj.parent is self._block)
            ],
        )
        self.assertEqual(
            [id(obj) for obj in order],
            [
                id(obj)
                for obj in self._preorder
                if (obj.parent is None) or (obj.parent is self._block)
            ],
        )

    def test_child(self):
        for child in self._child_key:
            parent = child.parent
            self.assertTrue(parent is not None)
            self.assertTrue(id(child) in set(id(_c) for _c in self._children[parent]))
            self.assertIs(parent.child(self._child_key[child]), child)
            with self.assertRaises(KeyError):
                parent.child("_not_a_valid_child_key_")

    def test_children(self):
        for obj in self._children:
            self.assertTrue(isinstance(obj, ICategorizedObjectContainer))
            if isinstance(obj, IBlock):
                for child in obj.children():
                    self.assertTrue(child.parent is obj)
                # this first test makes failures a
                # little easier to debug
                self.assertEqual(
                    sorted(str(child) for child in obj.children()),
                    sorted(str(child) for child in self._children[obj]),
                )
                self.assertEqual(
                    set(id(child) for child in obj.children()),
                    set(id(child) for child in self._children[obj]),
                )
                # this first test makes failures a
                # little easier to debug
                self.assertEqual(
                    sorted(str(child) for child in obj.children(ctype=IBlock)),
                    sorted(
                        str(child)
                        for child in self._children[obj]
                        if child.ctype is IBlock
                    ),
                )
                self.assertEqual(
                    set(id(child) for child in obj.children(ctype=IBlock)),
                    set(
                        id(child)
                        for child in self._children[obj]
                        if child.ctype is IBlock
                    ),
                )
                # this first test makes failures a
                # little easier to debug
                self.assertEqual(
                    sorted(str(child) for child in obj.children(ctype=IVariable)),
                    sorted(
                        str(child)
                        for child in self._children[obj]
                        if child.ctype is IVariable
                    ),
                )
                self.assertEqual(
                    set(id(child) for child in obj.children(ctype=IVariable)),
                    set(
                        id(child)
                        for child in self._children[obj]
                        if child.ctype is IVariable
                    ),
                )
            elif isinstance(obj, ICategorizedObjectContainer):
                for child in obj.children():
                    self.assertTrue(child.parent is obj)
                self.assertEqual(
                    set(id(child) for child in obj.children()),
                    set(id(child) for child in self._children[obj]),
                )
            else:
                self.assertEqual(len(self._children[obj]), 0)

    def test_components_no_descend_active_None(self):
        for obj in self._components_no_descend:
            self.assertTrue(isinstance(obj, ICategorizedObjectContainer))
            self.assertTrue(isinstance(obj, IBlock))
            for c in obj.components(descend_into=False):
                self.assertTrue(_path_to_object_exists(obj, c))
            # test ctype=IBlock
            self.assertEqual(
                sorted(
                    str(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=None, descend_into=False
                    )
                ),
                sorted(str(_b) for _b in self._components_no_descend[obj][IBlock]),
            )
            self.assertEqual(
                set(
                    id(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=None, descend_into=False
                    )
                ),
                set(id(_b) for _b in self._components_no_descend[obj][IBlock]),
            )
            # test ctype=IVariable
            self.assertEqual(
                sorted(
                    str(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=None, descend_into=False
                    )
                ),
                sorted(str(_v) for _v in self._components_no_descend[obj][IVariable]),
            )
            self.assertEqual(
                set(
                    id(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=None, descend_into=False
                    )
                ),
                set(id(_v) for _v in self._components_no_descend[obj][IVariable]),
            )
            # test no ctype
            self.assertEqual(
                sorted(
                    str(_c) for _c in obj.components(active=None, descend_into=False)
                ),
                sorted(
                    str(_c)
                    for ctype in self._components_no_descend[obj]
                    for _c in self._components_no_descend[obj][ctype]
                ),
            )
            self.assertEqual(
                set(id(_c) for _c in obj.components(active=None, descend_into=False)),
                set(
                    id(_c)
                    for ctype in self._components_no_descend[obj]
                    for _c in self._components_no_descend[obj][ctype]
                ),
            )

    def test_components_no_descend_active_True(self):
        for obj in self._components_no_descend:
            self.assertTrue(isinstance(obj, ICategorizedObjectContainer))
            self.assertTrue(isinstance(obj, IBlock))
            # test ctype=IBlock
            self.assertEqual(
                sorted(
                    str(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=True, descend_into=False
                    )
                ),
                (
                    sorted(
                        str(_b)
                        for _b in self._components_no_descend[obj][IBlock]
                        if _b.active
                    )
                    if getattr(obj, 'active', True)
                    else []
                ),
            )
            self.assertEqual(
                set(
                    id(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=True, descend_into=False
                    )
                ),
                (
                    set(
                        id(_b)
                        for _b in self._components_no_descend[obj][IBlock]
                        if _b.active
                    )
                    if getattr(obj, 'active', True)
                    else set()
                ),
            )
            # test ctype=IVariable
            self.assertEqual(
                sorted(
                    str(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=True, descend_into=False
                    )
                ),
                (
                    sorted(
                        str(_v) for _v in self._components_no_descend[obj][IVariable]
                    )
                    if getattr(obj, 'active', True)
                    else []
                ),
            )
            self.assertEqual(
                set(
                    id(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=True, descend_into=False
                    )
                ),
                (
                    set(id(_v) for _v in self._components_no_descend[obj][IVariable])
                    if getattr(obj, 'active', True)
                    else set()
                ),
            )
            # test no ctype
            self.assertEqual(
                sorted(
                    str(_c) for _c in obj.components(active=True, descend_into=False)
                ),
                (
                    sorted(
                        str(_c)
                        for ctype in self._components_no_descend[obj]
                        for _c in self._components_no_descend[obj][ctype]
                        if getattr(_c, "active", True)
                    )
                    if getattr(obj, 'active', True)
                    else []
                ),
            )
            self.assertEqual(
                set(id(_c) for _c in obj.components(active=True, descend_into=False)),
                (
                    set(
                        id(_c)
                        for ctype in self._components_no_descend[obj]
                        for _c in self._components_no_descend[obj][ctype]
                        if getattr(_c, "active", True)
                    )
                    if getattr(obj, 'active', True)
                    else set()
                ),
            )

    def test_components_active_None(self):
        for obj in self._components:
            self.assertTrue(isinstance(obj, ICategorizedObjectContainer))
            self.assertTrue(isinstance(obj, IBlock))
            for c in obj.components(descend_into=True):
                self.assertTrue(_path_to_object_exists(obj, c))
            # test ctype=IBlock
            self.assertEqual(
                sorted(
                    str(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=None, descend_into=True
                    )
                ),
                sorted(str(_b) for _b in self._components[obj][IBlock]),
            )
            self.assertEqual(
                set(
                    id(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=None, descend_into=True
                    )
                ),
                set(id(_b) for _b in self._components[obj][IBlock]),
            )
            # test ctype=IVariable
            self.assertEqual(
                sorted(
                    str(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=None, descend_into=True
                    )
                ),
                sorted(str(_v) for _v in self._components[obj][IVariable]),
            )
            self.assertEqual(
                set(
                    id(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=None, descend_into=True
                    )
                ),
                set(id(_v) for _v in self._components[obj][IVariable]),
            )
            # test no ctype
            self.assertEqual(
                sorted(
                    str(_c) for _c in obj.components(active=None, descend_into=True)
                ),
                sorted(
                    str(_c)
                    for ctype in self._components[obj]
                    for _c in self._components[obj][ctype]
                ),
            )
            self.assertEqual(
                set(id(_c) for _c in obj.components(active=None, descend_into=True)),
                set(
                    id(_c)
                    for ctype in self._components[obj]
                    for _c in self._components[obj][ctype]
                ),
            )

    def test_components_active_True(self):
        for obj in self._components:
            self.assertTrue(isinstance(obj, ICategorizedObjectContainer))
            self.assertTrue(isinstance(obj, IBlock))
            # test ctype=IBlock
            self.assertEqual(
                sorted(
                    str(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=True, descend_into=True
                    )
                ),
                (
                    sorted(str(_b) for _b in self._components[obj][IBlock] if _b.active)
                    if getattr(obj, 'active', True)
                    else []
                ),
            )
            self.assertEqual(
                set(
                    id(_b)
                    for _b in obj.components(
                        ctype=IBlock, active=True, descend_into=True
                    )
                ),
                (
                    set(id(_b) for _b in self._components[obj][IBlock] if _b.active)
                    if getattr(obj, 'active', True)
                    else set()
                ),
            )
            # test ctype=IVariable
            self.assertEqual(
                sorted(
                    str(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=True, descend_into=True
                    )
                ),
                (
                    sorted(
                        str(_v)
                        for _v in self._components[obj][IVariable]
                        if _active_path_to_object_exists(obj, _v)
                    )
                    if getattr(obj, 'active', True)
                    else []
                ),
            )
            self.assertEqual(
                set(
                    id(_v)
                    for _v in obj.components(
                        ctype=IVariable, active=True, descend_into=True
                    )
                ),
                (
                    set(
                        id(_v)
                        for _v in self._components[obj][IVariable]
                        if _active_path_to_object_exists(obj, _v)
                    )
                    if getattr(obj, 'active', True)
                    else set()
                ),
            )
            # test no ctype
            self.assertEqual(
                sorted(
                    str(_c) for _c in obj.components(active=True, descend_into=True)
                ),
                (
                    sorted(
                        str(_c)
                        for ctype in self._components[obj]
                        for _c in self._components[obj][ctype]
                        if _active_path_to_object_exists(obj, _c)
                    )
                    if getattr(obj, 'active', True)
                    else []
                ),
            )
            self.assertEqual(
                set(id(_c) for _c in obj.components(active=True, descend_into=True)),
                (
                    set(
                        id(_c)
                        for ctype in self._components[obj]
                        for _c in self._components[obj][ctype]
                        if _active_path_to_object_exists(obj, _c)
                    )
                    if getattr(obj, 'active', True)
                    else set()
                ),
            )


class _Test_block(_Test_block_base):
    _do_clone = None

    @classmethod
    def setUpClass(cls):
        assert cls._do_clone is not None
        model = cls._block = block()
        model.v_1 = variable()
        model.vdict_1 = variable_dict()
        model.vdict_1[None] = variable()
        model.vlist_1 = variable_list()
        model.vlist_1.append(variable())
        model.vlist_1.append(variable())
        model.b_1 = block()
        model.b_1.v_2 = variable()
        model.b_1.b_2 = block()
        model.b_1.b_2.b_3 = block()
        model.b_1.b_2.v_3 = variable()
        model.b_1.b_2.vlist_3 = variable_list()
        model.b_1.b_2.vlist_3.append(variable())
        model.b_1.b_2.deactivate(shallow=False)
        model.bdict_1 = block_dict()
        model.blist_1 = block_list()
        model.blist_1.append(block())
        model.blist_1[0].v_2 = variable()
        model.blist_1[0].b_2 = block()

        if cls._do_clone:
            model = cls._block = model.clone()

        #
        # Manually encode the correct output
        # for tests in the base testing class
        #

        cls._preorder = [
            model,
            model.v_1,
            model.vdict_1,
            model.vdict_1[None],
            model.vlist_1,
            model.vlist_1[0],
            model.vlist_1[1],
            model.b_1,
            model.b_1.v_2,
            model.b_1.b_2,
            model.b_1.b_2.b_3,
            model.b_1.b_2.v_3,
            model.b_1.b_2.vlist_3,
            model.b_1.b_2.vlist_3[0],
            model.bdict_1,
            model.blist_1,
            model.blist_1[0],
            model.blist_1[0].v_2,
            model.blist_1[0].b_2,
        ]

        cls._names = ComponentMap()
        cls._names[model.v_1] = "v_1"
        cls._names[model.vdict_1] = "vdict_1"
        cls._names[model.vdict_1[None]] = "vdict_1[None]"
        cls._names[model.vlist_1] = "vlist_1"
        cls._names[model.vlist_1[0]] = "vlist_1[0]"
        cls._names[model.vlist_1[1]] = "vlist_1[1]"
        cls._names[model.b_1] = "b_1"
        cls._names[model.b_1.v_2] = "b_1.v_2"
        cls._names[model.b_1.b_2] = "b_1.b_2"
        cls._names[model.b_1.b_2.b_3] = "b_1.b_2.b_3"
        cls._names[model.b_1.b_2.v_3] = "b_1.b_2.v_3"
        cls._names[model.b_1.b_2.vlist_3] = "b_1.b_2.vlist_3"
        cls._names[model.b_1.b_2.vlist_3[0]] = "b_1.b_2.vlist_3[0]"
        cls._names[model.bdict_1] = "bdict_1"
        cls._names[model.blist_1] = "blist_1"
        cls._names[model.blist_1[0]] = "blist_1[0]"
        cls._names[model.blist_1[0].v_2] = "blist_1[0].v_2"
        cls._names[model.blist_1[0].b_2] = "blist_1[0].b_2"

        cls._children = ComponentMap()
        cls._children[model] = [
            model.v_1,
            model.vdict_1,
            model.vlist_1,
            model.b_1,
            model.bdict_1,
            model.blist_1,
        ]
        cls._children[model.vdict_1] = [model.vdict_1[None]]
        cls._children[model.vlist_1] = [model.vlist_1[0], model.vlist_1[1]]
        cls._children[model.b_1] = [model.b_1.v_2, model.b_1.b_2]
        cls._children[model.b_1.b_2] = [
            model.b_1.b_2.v_3,
            model.b_1.b_2.vlist_3,
            model.b_1.b_2.b_3,
        ]
        cls._children[model.b_1.b_2.b_3] = []
        cls._children[model.b_1.b_2.vlist_3] = [model.b_1.b_2.vlist_3[0]]
        cls._children[model.bdict_1] = []
        cls._children[model.blist_1] = [model.blist_1[0]]
        cls._children[model.blist_1[0]] = [model.blist_1[0].v_2, model.blist_1[0].b_2]

        cls._child_key = ComponentMap()
        cls._child_key[model.v_1] = "v_1"
        cls._child_key[model.vdict_1] = "vdict_1"
        cls._child_key[model.vlist_1] = "vlist_1"
        cls._child_key[model.b_1] = "b_1"
        cls._child_key[model.bdict_1] = "bdict_1"
        cls._child_key[model.blist_1] = "blist_1"
        cls._child_key[model.vdict_1[None]] = None
        cls._child_key[model.vlist_1[0]] = 0
        cls._child_key[model.vlist_1[1]] = 1
        cls._child_key[model.b_1.v_2] = "v_2"
        cls._child_key[model.b_1.b_2] = "b_2"
        cls._child_key[model.b_1.b_2.b_3] = "b_3"
        cls._child_key[model.b_1.b_2.v_3] = "v_3"
        cls._child_key[model.b_1.b_2.vlist_3] = "vlist_3"
        cls._child_key[model.b_1.b_2.vlist_3[0]] = 0
        cls._child_key[model.blist_1[0]] = 0
        cls._child_key[model.blist_1[0].v_2] = "v_2"
        cls._child_key[model.blist_1[0].b_2] = "b_2"

        cls._components_no_descend = ComponentMap()
        cls._components_no_descend[model] = {}
        cls._components_no_descend[model][IVariable] = [
            model.v_1,
            model.vdict_1[None],
            model.vlist_1[0],
            model.vlist_1[1],
        ]
        cls._components_no_descend[model][IBlock] = [model.b_1, model.blist_1[0]]
        cls._components_no_descend[model.b_1] = {}
        cls._components_no_descend[model.b_1][IVariable] = [model.b_1.v_2]
        cls._components_no_descend[model.b_1][IBlock] = [model.b_1.b_2]
        cls._components_no_descend[model.b_1.b_2] = {}
        cls._components_no_descend[model.b_1.b_2][IVariable] = [
            model.b_1.b_2.v_3,
            model.b_1.b_2.vlist_3[0],
        ]
        cls._components_no_descend[model.b_1.b_2][IBlock] = [model.b_1.b_2.b_3]
        cls._components_no_descend[model.b_1.b_2.b_3] = {}
        cls._components_no_descend[model.b_1.b_2.b_3][IVariable] = []
        cls._components_no_descend[model.b_1.b_2.b_3][IBlock] = []
        cls._components_no_descend[model.blist_1[0]] = {}
        cls._components_no_descend[model.blist_1[0]][IVariable] = [model.blist_1[0].v_2]
        cls._components_no_descend[model.blist_1[0]][IBlock] = [model.blist_1[0].b_2]
        cls._components_no_descend[model.blist_1[0].b_2] = {}
        cls._components_no_descend[model.blist_1[0].b_2][IVariable] = []
        cls._components_no_descend[model.blist_1[0].b_2][IBlock] = []

        cls._components = ComponentMap()
        cls._components[model] = {}
        cls._components[model][IVariable] = [
            model.v_1,
            model.vdict_1[None],
            model.vlist_1[0],
            model.vlist_1[1],
            model.b_1.v_2,
            model.b_1.b_2.v_3,
            model.b_1.b_2.vlist_3[0],
            model.blist_1[0].v_2,
        ]
        cls._components[model][IBlock] = [
            model.b_1,
            model.blist_1[0],
            model.b_1.b_2,
            model.b_1.b_2.b_3,
            model.blist_1[0].b_2,
        ]
        cls._components[model.b_1] = {}
        cls._components[model.b_1][IVariable] = [
            model.b_1.v_2,
            model.b_1.b_2.v_3,
            model.b_1.b_2.vlist_3[0],
        ]
        cls._components[model.b_1][IBlock] = [model.b_1.b_2, model.b_1.b_2.b_3]
        cls._components[model.b_1.b_2] = {}
        cls._components[model.b_1.b_2][IVariable] = [
            model.b_1.b_2.v_3,
            model.b_1.b_2.vlist_3[0],
        ]
        cls._components[model.b_1.b_2][IBlock] = [model.b_1.b_2.b_3]
        cls._components[model.b_1.b_2.b_3] = {}
        cls._components[model.b_1.b_2.b_3][IVariable] = []
        cls._components[model.b_1.b_2.b_3][IBlock] = []
        cls._components[model.blist_1[0]] = {}
        cls._components[model.blist_1[0]][IVariable] = [model.blist_1[0].v_2]
        cls._components[model.blist_1[0]][IBlock] = [model.blist_1[0].b_2]
        cls._components[model.blist_1[0].b_2] = {}
        cls._components[model.blist_1[0].b_2][IVariable] = []
        cls._components[model.blist_1[0].b_2][IBlock] = []

        cls._blocks_no_descend = ComponentMap()
        for obj in cls._components_no_descend:
            cls._blocks_no_descend[obj] = [obj] + cls._components_no_descend[obj][
                IBlock
            ]

        cls._blocks = ComponentMap()
        for obj in cls._components:
            cls._blocks[obj] = [obj] + cls._components[obj][IBlock]

    def test_init(self):
        b = block()
        self.assertTrue(b.parent is None)
        self.assertEqual(b.ctype, IBlock)

    def test_type(self):
        b = block()
        self.assertTrue(isinstance(b, ICategorizedObject))
        self.assertTrue(isinstance(b, ICategorizedObjectContainer))
        self.assertTrue(isinstance(b, IHeterogeneousContainer))
        self.assertTrue(isinstance(b, IBlock))

    def test_overwrite(self):
        b = block()
        v = b.v = variable()
        self.assertIs(v.parent, b)
        b.v = variable()
        self.assertTrue(v.parent is None)

        # the same component can overwrite itself
        b = block()
        v = b.v = variable()
        self.assertIs(v.parent, b)
        b.v = v
        self.assertTrue(v.parent is b)

        b = block()
        c = b.c = constraint()
        self.assertIs(c.parent, b)
        b.c = constraint()
        self.assertTrue(c.parent is None)

        # the same component can overwrite itself
        b = block()
        c = b.c = constraint()
        self.assertIs(c.parent, b)
        b.c = c
        self.assertTrue(c.parent is b)

        b = block()
        v = b.v = variable()
        self.assertIs(v.parent, b)
        b.v = constraint()
        self.assertTrue(v.parent is None)

        b = block()
        c = b.c = variable()
        self.assertIs(c.parent, b)
        b.c = variable()
        self.assertTrue(c.parent is None)

    def test_already_has_parent(self):
        b1 = block()
        v = b1.v = variable()
        b2 = block()
        with self.assertRaises(ValueError):
            b2.v = v
        self.assertTrue(v.parent is b1)
        del b1.v
        b2.v = v
        self.assertTrue(v.parent is b2)

    def test_delattr(self):
        b = block()
        with self.assertRaises(AttributeError):
            del b.not_an_attribute
        c = b.b = block()
        self.assertIs(c.parent, b)
        del b.b
        self.assertIs(c.parent, None)
        b.b = c
        self.assertIs(c.parent, b)
        b.x = 2
        self.assertTrue(hasattr(b, 'x'))
        self.assertEqual(b.x, 2)
        del b.x
        self.assertTrue(not hasattr(b, 'x'))

    def test_collect_ctypes_small_block_storage(self):
        b = block()
        self.assertEqual(b.collect_ctypes(active=None), set())
        self.assertEqual(b.collect_ctypes(), set())
        self.assertEqual(b.collect_ctypes(active=True), set())
        b.x = variable()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable]))
        self.assertEqual(b.collect_ctypes(), set([IVariable]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable]))
        b.y = constraint()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable, IConstraint]))
        self.assertEqual(b.collect_ctypes(), set([IVariable, IConstraint]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable, IConstraint]))
        b.y.deactivate()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable, IConstraint]))
        self.assertEqual(b.collect_ctypes(), set([IVariable]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable]))
        B = block()
        B.b = b
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=None), set([IBlock])
        )
        self.assertEqual(B.collect_ctypes(descend_into=False), set([IBlock]))
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=True), set([IBlock])
        )
        self.assertEqual(
            B.collect_ctypes(active=None), set([IBlock, IVariable, IConstraint])
        )
        self.assertEqual(B.collect_ctypes(), set([IBlock, IVariable]))
        self.assertEqual(B.collect_ctypes(active=True), set([IBlock, IVariable]))
        b.deactivate()
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=None), set([IBlock])
        )
        self.assertEqual(B.collect_ctypes(descend_into=False), set([]))
        self.assertEqual(B.collect_ctypes(descend_into=False, active=True), set([]))
        self.assertEqual(
            B.collect_ctypes(active=None), set([IBlock, IVariable, IConstraint])
        )
        self.assertEqual(B.collect_ctypes(), set([]))
        self.assertEqual(B.collect_ctypes(active=True), set([]))
        B.x = variable()
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=None), set([IBlock, IVariable])
        )
        self.assertEqual(B.collect_ctypes(descend_into=False), set([IVariable]))
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=True), set([IVariable])
        )
        self.assertEqual(
            B.collect_ctypes(active=None), set([IBlock, IVariable, IConstraint])
        )
        self.assertEqual(B.collect_ctypes(), set([IVariable]))
        self.assertEqual(B.collect_ctypes(active=True), set([IVariable]))
        del b.y
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable]))
        self.assertEqual(b.collect_ctypes(), set([]))
        self.assertEqual(b.collect_ctypes(active=True), set([]))
        b.activate()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable]))
        self.assertEqual(b.collect_ctypes(), set([IVariable]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable]))

        del b.x
        self.assertEqual(b.collect_ctypes(), set())

    def test_collect_ctypes_large_block_storage(self):
        b = block()
        b._activate_large_storage_mode()
        self.assertEqual(b.collect_ctypes(active=None), set())
        self.assertEqual(b.collect_ctypes(), set())
        self.assertEqual(b.collect_ctypes(active=True), set())
        b.x = variable()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable]))
        self.assertEqual(b.collect_ctypes(), set([IVariable]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable]))
        b.y = constraint()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable, IConstraint]))
        self.assertEqual(b.collect_ctypes(), set([IVariable, IConstraint]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable, IConstraint]))
        b.y.deactivate()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable, IConstraint]))
        self.assertEqual(b.collect_ctypes(), set([IVariable]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable]))
        B = block()
        b._activate_large_storage_mode()
        B.b = b
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=None), set([IBlock])
        )
        self.assertEqual(B.collect_ctypes(descend_into=False), set([IBlock]))
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=True), set([IBlock])
        )
        self.assertEqual(
            B.collect_ctypes(active=None), set([IBlock, IVariable, IConstraint])
        )
        self.assertEqual(B.collect_ctypes(), set([IBlock, IVariable]))
        self.assertEqual(B.collect_ctypes(active=True), set([IBlock, IVariable]))
        b.deactivate()
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=None), set([IBlock])
        )
        self.assertEqual(B.collect_ctypes(descend_into=False), set([]))
        self.assertEqual(B.collect_ctypes(descend_into=False, active=True), set([]))
        self.assertEqual(
            B.collect_ctypes(active=None), set([IBlock, IVariable, IConstraint])
        )
        self.assertEqual(B.collect_ctypes(), set([]))
        self.assertEqual(B.collect_ctypes(active=True), set([]))
        B.x = variable()
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=None), set([IBlock, IVariable])
        )
        self.assertEqual(B.collect_ctypes(descend_into=False), set([IVariable]))
        self.assertEqual(
            B.collect_ctypes(descend_into=False, active=True), set([IVariable])
        )
        self.assertEqual(
            B.collect_ctypes(active=None), set([IBlock, IVariable, IConstraint])
        )
        self.assertEqual(B.collect_ctypes(), set([IVariable]))
        self.assertEqual(B.collect_ctypes(active=True), set([IVariable]))
        del b.y
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable]))
        self.assertEqual(b.collect_ctypes(), set([]))
        self.assertEqual(b.collect_ctypes(active=True), set([]))
        b.activate()
        self.assertEqual(b.collect_ctypes(active=None), set([IVariable]))
        self.assertEqual(b.collect_ctypes(), set([IVariable]))
        self.assertEqual(b.collect_ctypes(active=True), set([IVariable]))

        del b.x
        self.assertEqual(b.collect_ctypes(), set())


class Test_block_noclone(_Test_block, unittest.TestCase):
    _do_clone = False


class Test_block_clone(_Test_block, unittest.TestCase):
    _do_clone = True


class _MyBlockBaseBase(block):
    __slots__ = ()

    def __init__(self):
        super(_MyBlockBaseBase, self).__init__()


class _MyBlockBase(_MyBlockBaseBase):
    __slots__ = ("b",)

    def __init__(self):
        super(_MyBlockBase, self).__init__()
        self.b = block()
        self.b.v = variable()


class _MyBlock(_MyBlockBase):
    # testing when a __dict__ might appear (no __slots__)
    def __init__(self):
        super(_MyBlock, self).__init__()
        self.b.blist = block_list()
        self.b.blist.append(block())
        self.v = variable()
        self.n = 2.0


class _Test_small_block(_Test_block_base):
    _do_clone = None

    @classmethod
    def setUpClass(cls):
        assert cls._do_clone is not None
        cls._myblock_type = _MyBlock
        model = cls._block = _MyBlock()

        if cls._do_clone:
            model = cls._block = model.clone()

        #
        # Manually encode the correct output
        # for tests in the base testing class
        #

        cls._preorder = [
            model,
            model.b,
            model.b.v,
            model.b.blist,
            model.b.blist[0],
            model.v,
        ]

        cls._names = ComponentMap()
        cls._names[model.b] = "b"
        cls._names[model.b.v] = "b.v"
        cls._names[model.b.blist] = "b.blist"
        cls._names[model.b.blist[0]] = "b.blist[0]"
        cls._names[model.v] = "v"

        cls._children = ComponentMap()
        cls._children[model] = [model.b, model.v]
        cls._children[model.b] = [model.b.v, model.b.blist]
        cls._children[model.b.blist] = [model.b.blist[0]]
        cls._children[model.b.blist[0]] = []

        cls._child_key = ComponentMap()
        cls._child_key[model.b] = "b"
        cls._child_key[model.b.v] = "v"
        cls._child_key[model.b.blist] = "blist"
        cls._child_key[model.b.blist[0]] = 0
        cls._child_key[model.v] = "v"

        cls._components_no_descend = ComponentMap()
        cls._components_no_descend[model] = {}
        cls._components_no_descend[model][IVariable] = [model.v]
        cls._components_no_descend[model][IBlock] = [model.b]
        cls._components_no_descend[model.b] = {}
        cls._components_no_descend[model.b][IVariable] = [model.b.v]
        cls._components_no_descend[model.b][IBlock] = [model.b.blist[0]]
        cls._components_no_descend[model.b.blist[0]] = {}
        cls._components_no_descend[model.b.blist[0]][IBlock] = []
        cls._components_no_descend[model.b.blist[0]][IVariable] = []

        cls._components = ComponentMap()
        cls._components[model] = {}
        cls._components[model][IVariable] = [model.v, model.b.v]
        cls._components[model][IBlock] = [model.b, model.b.blist[0]]
        cls._components[model.b] = {}
        cls._components[model.b][IVariable] = [model.b.v]
        cls._components[model.b][IBlock] = [model.b.blist[0]]
        cls._components[model.b.blist[0]] = {}
        cls._components[model.b.blist[0]][IBlock] = []
        cls._components[model.b.blist[0]][IVariable] = []

        cls._blocks_no_descend = ComponentMap()
        for obj in cls._components_no_descend:
            cls._blocks_no_descend[obj] = [obj] + cls._components_no_descend[obj][
                IBlock
            ]

        cls._blocks = ComponentMap()
        for obj in cls._components:
            cls._blocks[obj] = [obj] + cls._components[obj][IBlock]

    # override this test method on the base class
    def test_collect_ctypes(self):
        self.assertEqual(
            self._block.collect_ctypes(active=None), set([IBlock, IVariable])
        )
        self.assertEqual(self._block.collect_ctypes(), set([IBlock, IVariable]))
        self.assertEqual(
            self._block.collect_ctypes(active=True), set([IBlock, IVariable])
        )
        self.assertEqual(
            self._block.collect_ctypes(descend_into=False), set([IBlock, IVariable])
        )
        self.assertEqual(
            self._block.collect_ctypes(active=True, descend_into=False),
            set([IBlock, IVariable]),
        )

        self._block.b.deactivate()
        try:
            self.assertEqual(
                self._block.collect_ctypes(active=None), set([IBlock, IVariable])
            )
            self.assertEqual(self._block.collect_ctypes(), set([IVariable]))
            self.assertEqual(self._block.collect_ctypes(active=True), set([IVariable]))
            self.assertEqual(
                self._block.collect_ctypes(active=None, descend_into=False),
                set([IBlock, IVariable]),
            )
            self.assertEqual(
                self._block.collect_ctypes(descend_into=False), set([IVariable])
            )
            self.assertEqual(
                self._block.collect_ctypes(active=True, descend_into=False),
                set([IVariable]),
            )
        finally:
            # use a finally block in case there is a failure above
            self._block.b.activate()

    def test_customblock_delattr(self):
        b = _MyBlock()
        with self.assertRaises(AttributeError):
            del b.not_an_attribute
        c = b.b
        self.assertIs(c.parent, b)
        del b.b
        self.assertIs(c.parent, None)
        b.b = c
        self.assertIs(c.parent, b)
        b.x = 2
        self.assertTrue(hasattr(b, 'x'))
        self.assertEqual(b.x, 2)
        del b.x
        self.assertTrue(not hasattr(b, 'x'))

    def test_customblock_setattr(self):
        b = _MyBlockBase()
        self.assertIs(b.b.parent, b)
        self.assertIs(b.b.v.parent, b.b)
        with self.assertRaises(ValueError):
            b.b = b.b.v
        self.assertIs(b.b.parent, b)
        self.assertIs(b.b.v.parent, b.b)
        c = b.b
        self.assertIs(c.parent, b)
        # test the edge case in setattr
        b.b = c
        self.assertIs(c.parent, b)
        assert not hasattr(b, "g")
        with self.assertRaises(ValueError):
            b.g = b.b
        self.assertIs(b.b.parent, b)
        b.g = 1
        with self.assertRaises(ValueError):
            b.g = b.b
        self.assertEqual(b.g, 1)
        self.assertIs(b.b.parent, b)
        # test an overwrite
        b.b = block()
        self.assertIs(c.parent, None)
        self.assertIs(b.b.parent, b)

    def test_customblock__with_dict_setattr(self):
        # This one was given a __dict__
        b = _MyBlock()
        self.assertIs(b.v.parent, b)
        self.assertIs(b.b.parent, b)
        with self.assertRaises(ValueError):
            b.v = b.b
        self.assertIs(b.v.parent, b)
        self.assertIs(b.b.parent, b)
        b.not_an_attribute = 2
        v = b.v
        self.assertIs(v.parent, b)
        # test the edge case in setattr
        b.v = v
        self.assertIs(v.parent, b)
        # test an overwrite
        b.v = variable()
        self.assertIs(v.parent, None)
        self.assertIs(b.v.parent, b)

    def test_inactive_behavior(self):
        b = _MyBlock()
        b.deactivate()
        self.assertNotEqual(len(list(pmo.preorder_traversal(b, active=None))), 0)
        self.assertEqual(len(list(pmo.preorder_traversal(b))), 0)
        self.assertEqual(len(list(pmo.preorder_traversal(b, active=True))), 0)

        def descend(x):
            return True

        self.assertNotEqual(
            len(list(pmo.preorder_traversal(b, active=None, descend=descend))), 0
        )
        self.assertEqual(len(list(pmo.preorder_traversal(b, descend=descend))), 0)
        self.assertEqual(
            len(list(pmo.preorder_traversal(b, active=True, descend=descend))), 0
        )

        def descend(x):
            descend.seen.append(x)
            return x.active

        descend.seen = []
        self.assertEqual(
            len(list(pmo.preorder_traversal(b, active=None, descend=descend))), 1
        )
        self.assertEqual(len(descend.seen), 1)
        self.assertIs(descend.seen[0], b)

        self.assertNotEqual(len(list(b.components(active=None))), 0)
        self.assertEqual(len(list(b.components())), 0)
        self.assertEqual(len(list(b.components(active=True))), 0)

        self.assertNotEqual(len(list(pmo.generate_names(b, active=None))), 0)
        self.assertEqual(len(list(pmo.generate_names(b))), 0)
        self.assertEqual(len(list(pmo.generate_names(b, active=True))), 0)


class Test_small_block_noclone(_Test_small_block, unittest.TestCase):
    _do_clone = False


class Test_small_block_clone(_Test_small_block, unittest.TestCase):
    _do_clone = True


class Test_block_dict(_TestActiveDictContainerBase, unittest.TestCase):
    _container_type = block_dict
    _ctype_factory = lambda self: block()


class Test_block_tuple(_TestActiveTupleContainerBase, unittest.TestCase):
    _container_type = block_tuple
    _ctype_factory = lambda self: block()


class Test_block_list(_TestActiveListContainerBase, unittest.TestCase):
    _container_type = block_list
    _ctype_factory = lambda self: block()


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.contrib.piecewise.transform.piecewise_linear_transformation_base import (
    PiecewiseLinearTransformationBase,
)
from pyomo.core import Constraint, NonNegativeIntegers, Var
from pyomo.core.base import TransformationFactory
from pyomo.gdp import Disjunct, Disjunction


@TransformationFactory.register(
    'contrib.piecewise.reduced_inner_repn_gdp',
    doc="Convert piecewise-linear model to a GDP "
    "using an inner representation of the "
    "simplices that are the domains of the linear "
    "functions.",
)
class ReducedInnerRepresentationGDPTransformation(PiecewiseLinearTransformationBase):
    """
    Convert a model involving piecewise linear expressions into a GDP by
    representing the piecewise linear functions as Disjunctions where the
    simplices over which the linear functions are defined are represented
    in a reduced "inner" representation--as convex combinations of their extreme
    points. We refer to this as 'reduced' since we create only one multiplier
    for each extreme point in the union of the extreme points over all the
    simplices. Within the Disjuncts, we then enforce that all of the multipliers
    for extreme points not in the simplex are 0.

    This transformation can be called in one of two ways:
        1) The default, where 'descend_into_expressions' is False. This is
           more computationally efficient, but relies on the
           PiecewiseLinearFunctions being declared on the same Block in which
           they are used in Expressions (if you are hoping to maintain the
           original hierarchical structure of the model). In this mode,
           targets must be Blocks and/or PiecewiseLinearFunctions.
        2) With 'descend_into_expressions' True. This is less computationally
           efficient, but will respect hierarchical structure by finding
           uses of PiecewiseLinearFunctions in Constraint and Objective
           expressions and putting their transformed counterparts on the same
           parent Block as the component owning their parent expression. In
           this mode, targets must be Blocks, Constraints, and/or Objectives.
    """

    CONFIG = PiecewiseLinearTransformationBase.CONFIG()
    _transformation_name = 'pw_linear_reduced_inner_repn'

    def _transform_pw_linear_expr(self, pw_expr, pw_linear_func, transformation_block):
        transBlock = transformation_block.transformed_functions[
            len(transformation_block.transformed_functions)
        ]

        # get the PiecewiseLinearFunctionExpression
        dimension = pw_expr.nargs()
        transBlock.disjuncts = Disjunct(NonNegativeIntegers)
        substitute_var = transBlock.substitute_var = Var()
        pw_linear_func.map_transformation_var(pw_expr, substitute_var)
        substitute_var_lb = float('inf')
        substitute_var_ub = -float('inf')
        extreme_pts_by_simplex = {}
        linear_func_by_extreme_pt = {}
        # Save all the extreme points as sets since we will need to check set
        # containment to build the constraints fixing the multipliers to 0. We
        # can also build the data structure that will allow us to later build
        # the linear func expression
        for simplex, linear_func in zip(
            pw_linear_func._simplices, pw_linear_func._linear_functions
        ):
            extreme_pts = extreme_pts_by_simplex[simplex] = set()
            for idx in simplex:
                extreme_pts.add(idx)
                if idx not in linear_func_by_extreme_pt:
                    linear_func_by_extreme_pt[idx] = linear_func

            # We're going to want bounds on the substitute var, so we use
            # interval arithmetic to figure those out as we go.
            lb, ub = compute_bounds_on_expr(linear_func(*pw_expr.args))
            if lb is not None and lb < substitute_var_lb:
                substitute_var_lb = lb
            if ub is not None and ub > substitute_var_ub:
                substitute_var_ub = ub

        # set the bounds on the substitute var
        if substitute_var_lb < float('inf'):
            transBlock.substitute_var.setlb(substitute_var_lb)
        if substitute_var_ub > -float('inf'):
            transBlock.substitute_var.setub(substitute_var_ub)

        num_extreme_pts = len(pw_linear_func._points)
        # lambda[i] will be the multiplier for the extreme point with index i in
        # pw_linear_fun._points
        transBlock.lambdas = Var(range(num_extreme_pts), bounds=(0, 1))

        # Now that we have all of the extreme points, we can make the
        # disjunctive constraints
        for simplex in pw_linear_func._simplices:
            disj = transBlock.disjuncts[len(transBlock.disjuncts)]
            cons = disj.lambdas_zero_for_other_simplices = Constraint(
                NonNegativeIntegers
            )
            extreme_pts = extreme_pts_by_simplex[simplex]
            for i in range(num_extreme_pts):
                if i not in extreme_pts:
                    cons[len(cons)] = transBlock.lambdas[i] <= 0
        # Make the disjunction
        transBlock.pick_a_piece = Disjunction(
            expr=[d for d in transBlock.disjuncts.values()]
        )

        # Now we make the global constraints
        transBlock.convex_combo = Constraint(
            expr=sum(transBlock.lambdas[i] for i in range(num_extreme_pts)) == 1
        )
        transBlock.linear_func = Constraint(
            expr=sum(
                linear_func_by_extreme_pt[j](*pt) * transBlock.lambdas[j]
                for (j, pt) in enumerate(pw_linear_func._points)
            )
            == substitute_var
        )

        @transBlock.Constraint(range(dimension))
        def linear_combo(b, i):
            return pw_expr.args[i] == sum(
                pt[i] * transBlock.lambdas[j]
                for (j, pt) in enumerate(pw_linear_func._points)
            )

        return transBlock.substitute_var



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import sys

from pyomo.common.extensions import ExtensionBuilderFactory
from pyomo.scripting.pyomo_parser import add_subparser


class ExtensionBuilder:
    def create_parser(self, parser):
        return parser

    def call(self, args, unparsed):
        logger = logging.getLogger('pyomo.common')
        original_level = logger.level
        logger.setLevel(logging.INFO)
        try:
            return self._call_impl(args, unparsed, logger)
        finally:
            logger.setLevel(original_level)

    def _call_impl(self, args, unparsed, logger):
        results = []
        result_fmt = "[%s]  %s"
        returncode = 0
        for target in ExtensionBuilderFactory:
            try:
                ext = ExtensionBuilderFactory(target)
                if hasattr(ext, 'skip') and ext.skip():
                    result = 'SKIP'
                elif hasattr(ext, '__call__'):
                    ext(parallel=args.parallel)
                    result = ' OK '
                else:
                    # Extension was a simple function and already ran
                    result = ' OK '
            except SystemExit:
                _info = sys.exc_info()
                _cls = (
                    str(_info[0].__name__ if _info[0] is not None else "NoneType")
                    + ": "
                )
                logger.error(_cls + str(_info[1]))
                result = 'FAIL'
                returncode |= 2
            except:
                _info = sys.exc_info()
                _cls = (
                    str(_info[0].__name__ if _info[0] is not None else "NoneType")
                    + ": "
                )
                logger.error(_cls + str(_info[1]))
                result = 'FAIL'
                returncode |= 1
            results.append(result_fmt % (result, target))
        logger.info("Finished building Pyomo extensions.")
        logger.info(
            "The following extensions were built:\n    " + "\n    ".join(results)
        )
        return returncode


#
# Add a subparser for the download-extensions command
#
_extension_builder = ExtensionBuilder()
_parser = _extension_builder.create_parser(
    add_subparser(
        'build-extensions',
        func=_extension_builder.call,
        help='Build compiled extension modules',
        add_help=False,
        description='This builds all registered (compilable) extension modules',
    )
)

_parser.add_argument(
    '-j',
    '--parallel',
    action='store',
    type=int,
    dest='parallel',
    default=None,
    help="Build with this many processes/cores",
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import sys
import time

from pyomo.common.dependencies import mpi4py
from pyomo.contrib.benders.benders_cuts import BendersCutGenerator
import pyomo.environ as pyo

"""
To run this example:

mpirun -np 3 python farmer.py
"""


class Farmer:
    def __init__(self):
        self.crops = ['WHEAT', 'CORN', 'SUGAR_BEETS']
        self.total_acreage = 500
        self.PriceQuota = {'WHEAT': 100000.0, 'CORN': 100000.0, 'SUGAR_BEETS': 6000.0}
        self.SubQuotaSellingPrice = {'WHEAT': 170.0, 'CORN': 150.0, 'SUGAR_BEETS': 36.0}
        self.SuperQuotaSellingPrice = {'WHEAT': 0.0, 'CORN': 0.0, 'SUGAR_BEETS': 10.0}
        self.CattleFeedRequirement = {'WHEAT': 200.0, 'CORN': 240.0, 'SUGAR_BEETS': 0.0}
        self.PurchasePrice = {'WHEAT': 238.0, 'CORN': 210.0, 'SUGAR_BEETS': 100000.0}
        self.PlantingCostPerAcre = {'WHEAT': 150.0, 'CORN': 230.0, 'SUGAR_BEETS': 260.0}
        self.scenarios = [
            'BelowAverageScenario',
            'AverageScenario',
            'AboveAverageScenario',
        ]
        self.crop_yield = dict()
        self.crop_yield['BelowAverageScenario'] = {
            'WHEAT': 2.0,
            'CORN': 2.4,
            'SUGAR_BEETS': 16.0,
        }
        self.crop_yield['AverageScenario'] = {
            'WHEAT': 2.5,
            'CORN': 3.0,
            'SUGAR_BEETS': 20.0,
        }
        self.crop_yield['AboveAverageScenario'] = {
            'WHEAT': 3.0,
            'CORN': 3.6,
            'SUGAR_BEETS': 24.0,
        }
        self.scenario_probabilities = dict()
        self.scenario_probabilities['BelowAverageScenario'] = 0.3333
        self.scenario_probabilities['AverageScenario'] = 0.3334
        self.scenario_probabilities['AboveAverageScenario'] = 0.3333


def create_root(farmer):
    m = pyo.ConcreteModel()

    m.crops = pyo.Set(initialize=farmer.crops, ordered=True)
    m.scenarios = pyo.Set(initialize=farmer.scenarios, ordered=True)

    m.devoted_acreage = pyo.Var(m.crops, bounds=(0, farmer.total_acreage))
    m.eta = pyo.Var(m.scenarios)
    for s in m.scenarios:
        m.eta[s].setlb(-432000 * farmer.scenario_probabilities[s])

    m.total_acreage_con = pyo.Constraint(
        expr=sum(m.devoted_acreage.values()) <= farmer.total_acreage
    )

    m.obj = pyo.Objective(
        expr=sum(
            farmer.PlantingCostPerAcre[crop] * m.devoted_acreage[crop]
            for crop in m.crops
        )
        + sum(m.eta.values())
    )
    return m


def create_subproblem(root, farmer, scenario):
    m = pyo.ConcreteModel()

    m.crops = pyo.Set(initialize=farmer.crops, ordered=True)

    m.devoted_acreage = pyo.Var(m.crops)
    m.QuantitySubQuotaSold = pyo.Var(m.crops, bounds=(0.0, None))
    m.QuantitySuperQuotaSold = pyo.Var(m.crops, bounds=(0.0, None))
    m.QuantityPurchased = pyo.Var(m.crops, bounds=(0.0, None))

    def EnforceCattleFeedRequirement_rule(m, i):
        return (
            farmer.CattleFeedRequirement[i]
            <= (farmer.crop_yield[scenario][i] * m.devoted_acreage[i])
            + m.QuantityPurchased[i]
            - m.QuantitySubQuotaSold[i]
            - m.QuantitySuperQuotaSold[i]
        )

    m.EnforceCattleFeedRequirement = pyo.Constraint(
        m.crops, rule=EnforceCattleFeedRequirement_rule
    )

    def LimitAmountSold_rule(m, i):
        return (
            m.QuantitySubQuotaSold[i]
            + m.QuantitySuperQuotaSold[i]
            - (farmer.crop_yield[scenario][i] * m.devoted_acreage[i])
            <= 0.0
        )

    m.LimitAmountSold = pyo.Constraint(m.crops, rule=LimitAmountSold_rule)

    def EnforceQuotas_rule(m, i):
        return (0.0, m.QuantitySubQuotaSold[i], farmer.PriceQuota[i])

    m.EnforceQuotas = pyo.Constraint(m.crops, rule=EnforceQuotas_rule)

    obj_expr = sum(
        farmer.PurchasePrice[crop] * m.QuantityPurchased[crop] for crop in m.crops
    )
    obj_expr -= sum(
        farmer.SubQuotaSellingPrice[crop] * m.QuantitySubQuotaSold[crop]
        for crop in m.crops
    )
    obj_expr -= sum(
        farmer.SuperQuotaSellingPrice[crop] * m.QuantitySuperQuotaSold[crop]
        for crop in m.crops
    )
    m.obj = pyo.Objective(expr=farmer.scenario_probabilities[scenario] * obj_expr)

    complicating_vars_map = pyo.ComponentMap()
    for crop in m.crops:
        complicating_vars_map[root.devoted_acreage[crop]] = m.devoted_acreage[crop]

    return m, complicating_vars_map


def main():
    rank = mpi4py.MPI.COMM_WORLD.Get_rank()
    if rank != 0:
        sys.stdout = open(os.devnull, 'w')

    t0 = time.time()
    farmer = Farmer()
    m = create_root(farmer=farmer)
    root_vars = list(m.devoted_acreage.values())
    m.benders = BendersCutGenerator()
    m.benders.set_input(root_vars=root_vars, tol=1e-8)
    for s in farmer.scenarios:
        subproblem_fn_kwargs = dict()
        subproblem_fn_kwargs['root'] = m
        subproblem_fn_kwargs['farmer'] = farmer
        subproblem_fn_kwargs['scenario'] = s
        m.benders.add_subproblem(
            subproblem_fn=create_subproblem,
            subproblem_fn_kwargs=subproblem_fn_kwargs,
            root_eta=m.eta[s],
            subproblem_solver='gurobi_persistent',
        )
    opt = pyo.SolverFactory('gurobi_persistent')
    opt.set_instance(m)

    print(
        '{0:<15}{1:<15}{2:<15}{3:<15}{4:<15}'.format(
            '# Cuts', 'Corn', 'Sugar Beets', 'Wheat', 'Time'
        )
    )
    for i in range(30):
        res = opt.solve(tee=False, save_results=False)
        cuts_added = m.benders.generate_cut()
        for c in cuts_added:
            opt.add_constraint(c)
        print(
            '{0:<15}{1:<15.2f}{2:<15.2f}{3:<15.2f}{4:<15.2f}'.format(
                len(cuts_added),
                m.devoted_acreage['CORN'].value,
                m.devoted_acreage['SUGAR_BEETS'].value,
                m.devoted_acreage['WHEAT'].value,
                time.time() - t0,
            )
        )
        if len(cuts_added) == 0:
            break


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Part of this module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

import enum
import glob
import logging
import math
import os
import operator
import re
import subprocess
import sys
from io import StringIO

# Now, import the base unittest environment.  We will override things
# specifically later
from unittest import *
import unittest as _unittest

from pyomo.common.collections import Mapping, Sequence
from pyomo.common.dependencies import attempt_import, check_min_version, multiprocessing
from pyomo.common.errors import InvalidValueError
from pyomo.common.fileutils import import_file
from pyomo.common.log import LoggingIntercept, pyomo_formatter
from pyomo.common.tee import capture_output

from unittest import mock

# We defer this import so that we don't add a hard dependence on pytest.
# Note that importing test modules may cause this import to be resolved
# (and then enforce a strict dependence on pytest)
pytest, pytest_available = attempt_import('pytest')


def _defaultFormatter(msg, default):
    return msg or default


def _floatOrCall(val):
    """Cast the value to float, if that fails call it and then cast.

    This is an "augmented" version of float() to better support
    integration with Pyomo NumericValue objects: if the initial cast to
    float fails by throwing a TypeError (as non-constant NumericValue
    objects will), then it falls back on calling the object and
    returning that value cast to float.

    """
    try:
        return float(val)
    except (TypeError, InvalidValueError):
        pass
    try:
        return float(val())
    except (TypeError, InvalidValueError):
        pass
    try:
        return val.value
    except AttributeError:
        # likely a complex
        return val


def assertStructuredAlmostEqual(
    first,
    second,
    places=None,
    msg=None,
    delta=None,
    reltol=None,
    abstol=None,
    allow_second_superset=False,
    item_callback=_floatOrCall,
    exception=ValueError,
    formatter=_defaultFormatter,
):
    """Test that first and second are equal up to a tolerance

    This compares first and second using both an absolute (`abstol`) and
    relative (`reltol`) tolerance.  It will recursively descend into
    Sequence and Mapping containers (allowing for the relative
    comparison of structured data including lists and dicts).

    `places` and `delta` is supported for compatibility with
    assertAlmostEqual.  If `places` is supplied, `abstol` is
    computed as `10**-places`.  `delta` is an alias for `abstol`.

    If none of {`abstol`, `reltol`, `places`, `delta`} are specified,
    `reltol` defaults to 1e-7.

    If `allow_second_superset` is True, then:

      - only key/value pairs found in mappings in `first` are
        compared to `second` (allowing mappings in `second` to
        contain extra keys)

      - only values found in sequences in `first` are compared to
        `second` (allowing sequences in `second` to contain extra
        values)

    The relative error is computed for numerical values as

        `abs(first - second) / max(abs(first), abs(second))`

    only when `first != second` (thereby avoiding divide-by-zero errors).

    Items (entries other than Sequence / Mapping containers, matching
    strings, and items that satisfy `first is second`) are passed to the
    `item_callback` before testing equality and relative tolerances.

    Raises `exception` if `first` and `second` are not equal within
    tolerance.

    Parameters
    ----------
    first :
        the first value to compare

    second :
        the second value to compare

    places : int
        `first` and `second` are considered equivalent if their
        difference is between `places` decimal places; equivalent to
        `abstol = 10**-places` (included for compatibility with
        assertAlmostEqual)

    msg : str
        the message to raise on failure

    delta : float
        alias for `abstol`

    abstol : float
        the absolute tolerance.  `first` and `second` are considered
        equivalent if their absolute difference is less than `abstol`

    reltol : float
        the relative tolerance.  `first` and `second` are considered
        equivalent if their absolute difference divided by the
        largest of `first` and `second` is less than `reltol`

    allow_second_superset : bool
        If True, then extra entries in containers found on second
        will not trigger a failure.

    item_callback : function
        items (other than Sequence / Mapping containers, matching
        strings, and items satisfying `is`) are passed to this callback
        to generate the (nominally floating point) value to use for
        comparison.

    exception : Exception
        exception to raise when `first` is not 'almost equal' to `second`.

    formatter : function
        callback for generating the final failure message (for
        compatibility with unittest)

    """
    if sum(1 for _ in (places, delta, abstol) if _ is not None) > 1:
        raise ValueError("Cannot specify more than one of {places, delta, abstol}")

    if places is not None:
        abstol = 10 ** (-places)
    if delta is not None:
        abstol = delta
    if abstol is None and reltol is None:
        reltol = 10**-7

    fail = None
    try:
        _assertStructuredAlmostEqual(
            first,
            second,
            abstol,
            reltol,
            not allow_second_superset,
            item_callback,
            exception,
        )
    except exception as e:
        fail = formatter(
            msg,
            "%s\n    Found when comparing with tolerance "
            "(abs=%s, rel=%s):\n"
            "        first=%s\n        second=%s"
            % (
                str(e),
                abstol,
                reltol,
                _unittest.case.safe_repr(first),
                _unittest.case.safe_repr(second),
            ),
        )

    if fail:
        raise exception(fail)


def _assertStructuredAlmostEqual(
    first, second, abstol, reltol, exact, item_callback, exception
):
    """Recursive implementation of assertStructuredAlmostEqual"""

    args = (first, second)
    f, s = args
    if all(isinstance(_, Mapping) for _ in args):
        if exact and len(first) != len(second):
            raise exception(
                "mappings are different sizes (%s != %s)" % (len(first), len(second))
            )
        for key in first:
            if key not in second:
                raise exception(
                    "key (%s) from first not found in second"
                    % (_unittest.case.safe_repr(key),)
                )
            try:
                _assertStructuredAlmostEqual(
                    first[key],
                    second[key],
                    abstol,
                    reltol,
                    exact,
                    item_callback,
                    exception,
                )
            except exception as e:
                raise exception(
                    "%s\n    Found when comparing key %s"
                    % (str(e), _unittest.case.safe_repr(key))
                )
        return  # PASS!

    elif any(isinstance(_, str) for _ in args):
        if first == second:
            return  # PASS!

    elif all(isinstance(_, Sequence) for _ in args):
        # Note that Sequence includes strings
        if exact and len(first) != len(second):
            raise exception(
                "sequences are different sizes (%s != %s)" % (len(first), len(second))
            )
        for i, (f, s) in enumerate(zip(first, second)):
            try:
                _assertStructuredAlmostEqual(
                    f, s, abstol, reltol, exact, item_callback, exception
                )
            except exception as e:
                raise exception("%s\n    Found at position %s" % (str(e), i))
        return  # PASS!

    else:
        # Catch things like None, which may cause problems for the
        # item_callback [like float(None)])
        #
        # Test `is` and `==`, but this is not necessarily fatal: we will
        # continue and allow the item_callback to potentially convert
        # the values to be comparable.
        try:
            if first is second or first == second:
                return  # PASS!
        except:
            pass
        try:
            f = item_callback(first)
            s = item_callback(second)
            if f == s:
                return
            diff = abs(f - s)
            if abstol is not None and diff <= abstol:
                return  # PASS!
            if reltol is not None and diff / max(abs(f), abs(s)) <= reltol:
                return  # PASS!
            if math.isnan(f) and math.isnan(s):
                return  # PASS! (we will treat NaN as equal)
        except:
            pass

    msg = "%s !~= %s" % (
        _unittest.case.safe_repr(first),
        _unittest.case.safe_repr(second),
    )
    if f is not first or s is not second:
        msg = "%s !~= %s (%s)" % (
            _unittest.case.safe_repr(f),
            _unittest.case.safe_repr(s),
            msg,
        )
    raise exception(msg)


def _runner(pipe, qualname):
    "Utility wrapper for running functions, used by timeout()"
    resultType = _RunnerResult.call
    if pipe in _runner.data:
        fcn, args, kwargs = _runner.data[pipe]
    elif isinstance(qualname, str):
        # Use unittest to instantiate the TestCase and run it
        resultType = _RunnerResult.unittest

        def fcn():
            suite = _unittest.TestLoader().loadTestsFromName(qualname)
            result = _unittest.TestResult()
            # Starting in Python 3.14, the default interface is
            # forkserver, so timeout will fall back on spawn (getting us
            # here).  Unfortunately, starting in pytest 9.0, if the
            # test is expecting to fail, pytest will completely suppress
            # recording the failure from the subTest, causing the main
            # test to unexpectedly succeed.  We will resolve this by
            # looking at the testmethod we just loaded and if we are
            # expecting a failure, then we will turn that off *in the
            # subTest* so that the result is properly propagated.
            for test in suite:
                test.__unittest_expecting_failure__ = False
                # Note: fetch the unbound function off the class and not
                # the bound method.
                func = getattr(test.__class__, test._testMethodName)
                if hasattr(func, '__unittest_expecting_failure__'):
                    delattr(func, '__unittest_expecting_failure__')
            # Now we can actually run the test (including all necessary
            # setUp/tearDown)
            suite.run(result)
            return result.errors + result.failures, result.skipped

        args = ()
        kwargs = {}
    else:
        qualname, fcn, args, kwargs = qualname
    _runner.data[qualname] = None
    try:
        with capture_output() as OUT:
            result = fcn(*args, **kwargs)
        pipe.send((resultType, result, OUT.getvalue()))
    except:
        import traceback

        etype, e, tb = sys.exc_info()
        if not isinstance(e, AssertionError):
            e = etype(
                "%s\nOriginal traceback:\n%s" % (e, ''.join(traceback.format_tb(tb)))
            )
        pipe.send((_RunnerResult.exception, e, OUT.getvalue()))
    finally:
        _runner.data.pop(qualname)


# Data structure for passing functions/arguments to the _runner
# without forcing them to be pickled / unpickled
_runner.data = {}


class _RunnerResult(enum.Enum):
    exception = 0
    call = 1
    unittest = 2


def timeout(seconds, require_fork=False, timeout_raises=TimeoutError):
    """Function decorator to timeout the decorated function.

    This decorator will wrap a function call with a timeout, returning
    the result of the wrapped function.  The timeout is implemented
    using multiprocessing to execute the function in a forked process.
    If the wrapped function raises an exception, then the exception will
    be re-raised in this process.  If the function times out, a
    :class:`TimeoutError` will be raised.

    Note that as this method uses multiprocessing, the wrapped function
    should NOT spawn any subprocesses.  The timeout is implemented using
    `multiprocessing.Process.terminate()`, which sends a SIGTERM signal
    to the subprocess.  Any spawned subprocesses are not collected and
    will be orphaned and left running.

    Parameters
    ----------
    seconds: float
        Number of seconds to wait before timing out the function

    require_fork: bool
        Require support of the 'fork' interface.  If not present,
        immediately raises unittest.SkipTest

    timeout_raises: Exception
        Exception class to raise in the event of a timeout

    Examples
    --------
    .. doctest::
       :skipif: multiprocessing.get_start_method() != 'fork'

       >>> import pyomo.common.unittest as unittest
       >>> @unittest.timeout(1)
       ... def test_function():
       ...     return 42
       >>> test_function()
       42

    .. doctest::
       :skipif: multiprocessing.get_start_method() != 'fork'

       >>> @unittest.timeout(0.01)
       ... def test_function():
       ...     while 1:
       ...         pass
       >>> test_function()
       Traceback (most recent call last):
           ...
       TimeoutError: test timed out after 0.01 seconds

    """
    import functools
    import queue

    def timeout_decorator(fcn):
        @functools.wraps(fcn)
        def test_timer(*args, **kwargs):
            qualname = '%s.%s' % (fcn.__module__, fcn.__qualname__)
            # If qualname is in the data dict, then we are in the child
            # process and are being asked to run the wrapped function.
            if qualname in _runner.data:
                return fcn(*args, **kwargs)
            # Parent process: spawn a subprocess to execute the wrapped
            # function and monitor for timeout
            if require_fork and multiprocessing.get_start_method() != 'fork':
                raise _unittest.SkipTest(
                    "timeout() requires unavailable fork interface"
                )

            pipe_recv, pipe_send = multiprocessing.Pipe(False)
            if multiprocessing.get_start_method() == 'fork':
                # Option 1: leverage fork if possible.  This minimizes
                # the reliance on serialization and ensures that the
                # wrapped function operates in the same environment.
                _runner.data[pipe_send] = (fcn, args, kwargs)
                runner_arg = qualname
            elif (
                args
                and fcn.__name__.startswith('test')
                and _unittest.case.TestCase in args[0].__class__.__mro__
            ):
                # Option 2: this is wrapping a unittest.  Re-run
                # unittest in the child process with this function as
                # the sole target.  This ensures that things like setUp
                # and tearDown are correctly called.
                runner_arg = qualname
            else:
                # Option 3: attempt to serialize the function and all
                # arguments and send them to the (spawned) child
                # process.  The wrapped function cannot count on any
                # environment configuration that it does not set up
                # itself.
                runner_arg = (qualname, test_timer, args, kwargs)
            test_proc = multiprocessing.Process(
                target=_runner, args=(pipe_send, runner_arg)
            )
            # Set daemon: if the parent process is killed, the child
            # process should be killed and collected.
            test_proc.daemon = True
            try:
                test_proc.start()
            except:
                if type(runner_arg) is tuple:
                    logging.getLogger(__name__).error(
                        "Exception raised spawning timeout() subprocess "
                        "on a platform that does not support 'fork'.  "
                        "It is likely that either the wrapped function or "
                        "one of its arguments is not serializable"
                    )
                raise
            try:
                if pipe_recv.poll(seconds):
                    resultType, result, stdout = pipe_recv.recv()
                else:
                    test_proc.terminate()
                    raise timeout_raises(
                        "test timed out after %s seconds" % (seconds,)
                    ) from None
            finally:
                _runner.data.pop(pipe_send, None)
            sys.stdout.write(stdout)
            test_proc.join()
            if resultType == _RunnerResult.call:
                return result
            elif resultType == _RunnerResult.unittest:
                for name, msg in result[0]:
                    with args[0].subTest(name):
                        raise args[0].failureException(msg)
                for name, msg in result[1]:
                    with args[0].subTest(name):
                        args[0].skipTest(msg)
            else:
                raise result

        return test_timer

    return timeout_decorator


class _AssertRaisesContext_NormalizeWhitespace(_unittest.case._AssertRaisesContext):
    def __exit__(self, exc_type, exc_value, tb):
        try:
            _save_re = self.expected_regex
            self.expected_regex = None
            if not super().__exit__(exc_type, exc_value, tb):
                return False
        finally:
            self.expected_regex = _save_re

        exc_value = re.sub(r'(?s)\s+', ' ', str(exc_value))
        if not _save_re.search(exc_value):
            self._raiseFailure(
                '"{}" does not match "{}"'.format(_save_re.pattern, exc_value)
            )
        return True


class TestCase(_unittest.TestCase):
    """A Pyomo-specific class whose instances are single test cases.

    This class derives from unittest.TestCase and provides the following
    additional functionality:

    * additional assertions:
       - :py:meth:`~TestCase.assertStructuredAlmostEqual`
       - :py:meth:`assertExpressionsEqual`
       - :py:meth:`assertExpressionsStructurallyEqual`

    * updated assertions:
       - :py:meth:`assertRaisesRegex`

    :py:class:`unittest.TestCase` documentation
    -------------------------------------------
    """

    # Note that the current unittest.TestCase documentation generates
    # sphinx warnings.  We will clean up that documentation to suppress
    # the warnings.
    __doc__ += (
        re.sub(
            r'^( +)(\* +[^:]+:) *',
            r'\n\1\2\n\1    ',
            _unittest.TestCase.__doc__.rstrip(),
            flags=re.M,
        )
        + "\n\n"
    )

    # By default, we always want to spend the time to create the full
    # diff of the test reault and the baseline
    maxDiff = None

    def assertStructuredAlmostEqual(
        self,
        first,
        second,
        places=None,
        msg=None,
        delta=None,
        reltol=None,
        abstol=None,
        allow_second_superset=False,
        item_callback=_floatOrCall,
    ):
        # Note: __doc__ copied from assertStructuredAlmostEqual below
        #
        assertStructuredAlmostEqual(
            first=first,
            second=second,
            places=places,
            msg=msg,
            delta=delta,
            reltol=reltol,
            abstol=abstol,
            allow_second_superset=allow_second_superset,
            item_callback=item_callback,
            exception=self.failureException,
            formatter=self._formatMessage,
        )

    def assertRaisesRegex(self, expected_exception, expected_regex, *args, **kwargs):
        """Asserts that the message in a raised exception matches a regex.

        This is a light weight wrapper around
        :py:meth:`unittest.TestCase.assertRaisesRegex` that adds
        handling of a `normalize_whitespace` keyword argument that
        normalizes all consecutive whitespace in the exception message
        to a single space before checking the regular expression.

        Parameters
        ----------
        expected_exception : Exception
            Exception class expected to be raised.

        expected_regex : `re.Pattern` or str
            Regular expression expected to be found in error message.

        *args :
            Function to be called and extra positional args.

        **kwargs :
            Extra keyword args.

        msg : str
            Optional message used in case of failure. Can only be used
            when assertRaisesRegex is used as a context manager.

        normalize_whitespace : bool, default=False
            If True, collapses consecutive whitespace (including
            newlines) into a single space before checking against the
            regular expression

        """
        normalize_whitespace = kwargs.pop('normalize_whitespace', False)
        if normalize_whitespace:
            contextClass = _AssertRaisesContext_NormalizeWhitespace
        else:
            contextClass = _unittest.case._AssertRaisesContext
        context = contextClass(expected_exception, self, expected_regex)
        return context.handle('assertRaisesRegex', args, kwargs)

    def assertExpressionsEqual(self, a, b, include_named_exprs=True, places=None):
        """Assert that two Pyomo expressions are equal.

        This converts the expressions `a` and `b` into prefix notation
        and then compares the resulting lists.  All nodes in the tree
        are compared using py:meth:`assertEqual` (or
        py:meth:`assertAlmostEqual`)

        Parameters
        ----------
        a: ExpressionBase or native type

        b: ExpressionBase or native type

        include_named_exprs : bool
            If True (the default), the comparison expands all named
            expressions when generating the prefix notation

        places : float
            Number of decimal places required for equality of floating
            point numbers in the expression. If None (the default), the
            expressions must be exactly equal.

        """
        from pyomo.core.expr.compare import assertExpressionsEqual

        return assertExpressionsEqual(self, a, b, include_named_exprs, places)

    def assertExpressionsStructurallyEqual(
        self, a, b, include_named_exprs=True, places=None
    ):
        """Assert that two Pyomo expressions are structurally equal.

        This converts the expressions `a` and `b` into prefix notation
        and then compares the resulting lists.  Operators and
        (non-native type) leaf nodes in the prefix representation are
        converted to strings before comparing (so that things like
        variables can be compared across clones or pickles)

        Parameters
        ----------
        a: ExpressionBase or native type

        b: ExpressionBase or native type

        include_named_exprs: bool
            If True (the default), the comparison expands all named
            expressions when generating the prefix notation

        places: float
            Number of decimal places required for equality of floating
            point numbers in the expression. If None (the default), the
            expressions must be exactly equal.

        """
        from pyomo.core.expr.compare import assertExpressionsStructurallyEqual

        return assertExpressionsStructurallyEqual(
            self, a, b, include_named_exprs, places
        )


TestCase.assertStructuredAlmostEqual.__doc__ = re.sub(
    'exception :.*', '', assertStructuredAlmostEqual.__doc__, flags=re.S
)


class BaselineTestDriver:
    """Generic driver for performing baseline tests in bulk

    This test driver was originally crafted for testing the examples in
    the Pyomo Book, and has since been generalized to reuse in testing
    ".. literalinclude:" examples from the Online Docs.

    We expect that consumers of this class will derive from both this
    class and `pyomo.common.unittest.TestCase`, and then use
    `parameterized` to declare tests that call either the
    :py:meth:`python_test_driver` or :py:meth:`shell_test_driver`
    methods.

    Note that derived classes must declare two class attributes:

    Class Attributes
    ----------------
    solver_dependencies: Dict[str, List[str]]

        maps the test name to a list of required solvers.  If any solver
        is not available, then the test will be skipped.

    package_dependencies: Dict[str, List[str]]

        maps the test name to a list of required modules.  If any module
        is not available, then the test will be skipped.

    """

    @staticmethod
    def custom_name_func(test_func, test_num, test_params):
        func_name = test_func.__name__
        return "test_%s_%s" % (test_params.args[0], func_name[-2:])

    def __init__(self, test):
        # Finalize the class, if necessary...
        if getattr(self.__class__, 'solver_available', None) is None:
            self.initialize_dependencies()
        super().__init__(test)

    def initialize_dependencies(self):
        # Note: as a rule, pyomo.common is not allowed to import from
        # the rest of Pyomo.  we permit it here because a) this is not
        # at module scope, and b) there is really no better / more
        # logical place in pyomo to put this code.
        from pyomo.opt import check_available_solvers

        cls = self.__class__
        #
        # Initialize the availability data
        #
        solvers_used = set(sum(list(cls.solver_dependencies.values()), []))
        available_solvers = check_available_solvers(*solvers_used)
        cls.solver_available = {
            solver_: (solver_ in available_solvers) for solver_ in solvers_used
        }

        cls.package_available = {}
        cls.package_modules = {}
        packages_used = set(sum(list(cls.package_dependencies.values()), []))
        for package_ in packages_used:
            pack, pack_avail = attempt_import(package_, defer_import=False)
            cls.package_available[package_] = pack_avail
            cls.package_modules[package_] = pack

    @classmethod
    def _find_tests(cls, test_dirs, pattern):
        test_tuples = []
        for testdir in test_dirs:
            # Find all pattern files in the test directory and any immediate
            # sub-directories
            for fname in list(glob.glob(os.path.join(testdir, pattern))) + list(
                glob.glob(os.path.join(testdir, '*', pattern))
            ):
                test_file = os.path.abspath(fname)
                bname = os.path.basename(test_file)
                dir_ = os.path.dirname(test_file)
                name = os.path.splitext(bname)[0]
                tname = os.path.basename(dir_) + '_' + name

                suffix = None
                # Look for txt and yml file names matching py file names. Add
                # a test for any found
                for suffix_ in ['.txt', '.yml']:
                    if os.path.exists(os.path.join(dir_, name + suffix_)):
                        suffix = suffix_
                        break
                if suffix is not None:
                    tname = tname.replace('-', '_')
                    tname = tname.replace('.', '_')

                    # Create list of tuples with (test_name, test_file, baseline_file)
                    test_tuples.append(
                        (tname, test_file, os.path.join(dir_, name + suffix))
                    )

        # Ensure a deterministic test ordering
        test_tuples.sort()
        return test_tuples

    @classmethod
    def gather_tests(cls, test_dirs):
        # Find all .sh files in the test directories
        sh_test_tuples = cls._find_tests(test_dirs, '*.sh')

        # Find all .py files in the test directories
        py_test_tuples = cls._find_tests(test_dirs, '*.py')

        # If there is both a .py and a .sh, defer to the sh
        sh_files = set(map(operator.itemgetter(1), sh_test_tuples))
        py_test_tuples = list(
            filter(lambda t: t[1][:-3] + '.sh' not in sh_files, py_test_tuples)
        )

        return py_test_tuples, sh_test_tuples

    def check_skip(self, name):
        """
        Return a boolean if the test should be skipped
        """

        if name in self.solver_dependencies:
            solvers_ = self.solver_dependencies[name]
            if not all([self.solver_available[i] for i in solvers_]):
                # Skip the test because a solver is not available
                _missing = []
                for i in solvers_:
                    if not self.solver_available[i]:
                        _missing.append(i)
                return "Solver%s %s %s not available" % (
                    's' if len(_missing) > 1 else '',
                    ", ".join(_missing),
                    'are' if len(_missing) > 1 else 'is',
                )

        if name in self.package_dependencies:
            packages_ = self.package_dependencies[name]
            if not all([self.package_available[i] for i in packages_]):
                # Skip the test because a package is not available
                _missing = []
                for i in packages_:
                    if not self.package_available[i]:
                        _missing.append(i)
                return "Package%s %s %s not available" % (
                    's' if len(_missing) > 1 else '',
                    ", ".join(_missing),
                    'are' if len(_missing) > 1 else 'is',
                )

            # This is a hack, xlrd dropped support for .xlsx files in 2.0.1 which
            # causes problems with older versions of Pandas<=1.1.5 so skipping
            # tests requiring both these packages when incompatible versions are found
            if (
                'pandas' in self.package_dependencies[name]
                and 'xlrd' in self.package_dependencies[name]
            ):
                if check_min_version(
                    self.package_modules['xlrd'], '2.0.1'
                ) and not check_min_version(self.package_modules['pandas'], '1.1.6'):
                    return "Incompatible versions of xlrd and pandas"

        return False

    def filter_fcn(self, line):
        """
        Ignore certain text when comparing output with baseline
        """
        for field in (
            '[',
            'password:',
            'http:',
            'Job ',
            'Importing module',
            'Function',
            'File',
            'Matplotlib',
            'Memory:',
            '-------',
            '=======',
            '    ^',
        ):
            if line.startswith(field):
                return True
        for field in (
            'Total CPU',
            'Ipopt',
            'license',
            #'Status: optimal',
            #'Status: feasible',
            'time:',
            'Time:',
            'with format cpxlp',
            'usermodel = <module',
            'execution time=',
            'Solver results file:',
            'TokenServer',
            # next 6 patterns ignore entries in pstats reports:
            'function calls',
            'List reduced',
            '.py:',  # timing/profiling output
            ' {built-in method',
            ' {method',
            ' {pyomo.core.expr.numvalue.as_numeric}',
            ' {gurobipy.',
        ):
            if field in line:
                return True
        return False

    def filter_file_contents(self, lines, abstol=None):
        _numpy_scalar_re = re.compile(r'np.(int|float)\d+\(([^\)]+)\)')
        filtered = []
        deprecated = None
        for line in lines:
            # Ignore all deprecation warnings
            if line.startswith('WARNING: DEPRECATED:'):
                deprecated = ''
            if deprecated is not None:
                deprecated += line
                if re.search(r'\(called\s+from[^)]+\)', deprecated):
                    deprecated = None
                continue

            if not line or self.filter_fcn(line):
                continue

            # Strip off beginning of lines giving time in seconds
            # Needed for the performance chapter tests
            if "seconds" in line:
                s = line.find("seconds") + 7
                line = line[s:]

            item_list = []
            items = line.strip().split()
            for i in items:
                # Split up lists, dicts, and sets
                while i and i[0] in '[{':
                    item_list.append(i[0])
                    i = i[1:]
                tail = []
                while i and i[-1] in ',:]}':
                    tail.append(i[-1])
                    i = i[:-1]

                # A few substitutions to get tests passing on pypy3
                if ".inf" in i:
                    i = i.replace(".inf", "inf")
                if "null" in i:
                    i = i.replace("null", "None")

                try:
                    # Numpy 2.x changed the repr for scalars.  Convert
                    # the new scalar reprs back to the original (which
                    # were indistinguishable from python floats/ints)
                    np_match = _numpy_scalar_re.match(i)
                    if np_match:
                        item_list.append(float(np_match.group(2)))
                    else:
                        item_list.append(float(i))
                except:
                    item_list.append(i)
                if tail:
                    tail.reverse()
                    item_list.extend(tail)

            # We can get printed results objects where the baseline is
            # exactly 0 (and omitted) and the test is slightly non-zero.
            # We will look for the pattern of values printed from
            # results objects and remote them if they are within
            # tolerance of 0
            if (
                len(item_list) == 3
                and item_list[0] == 'Value'
                and item_list[1] == ':'
                and type(item_list[2]) is float
                and abs(item_list[2]) < (abstol or 0)
                and len(filtered[-1]) == 2
                and filtered[-1][1] == ':'
            ):
                filtered.pop()
            else:
                filtered.append(item_list)

        return filtered

    def compare_baseline(self, test_output, baseline, abstol=1e-6, reltol=1e-8):
        # Filter files independently and then compare filtered contents
        out_filtered = self.filter_file_contents(
            test_output.strip().split('\n'), abstol
        )
        base_filtered = self.filter_file_contents(baseline.strip().split('\n'), abstol)

        try:
            self.assertStructuredAlmostEqual(
                out_filtered,
                base_filtered,
                abstol=abstol,
                reltol=reltol,
                allow_second_superset=False,
            )
            return True
        except self.failureException:
            # Print helpful information when file comparison fails
            print('---------------------------------')
            print('BASELINE FILE')
            print('---------------------------------')
            print(baseline)
            print('=================================')
            print('---------------------------------')
            print('TEST OUTPUT FILE')
            print('---------------------------------')
            print(test_output)
            raise

    def python_test_driver(self, tname, test_file, base_file):
        bname = os.path.basename(test_file)
        _dir = os.path.dirname(test_file)

        skip_msg = self.check_skip('test_' + tname)
        if skip_msg:
            raise _unittest.SkipTest(skip_msg)

        with open(base_file, 'r') as FILE:
            baseline = FILE.read()

        cwd = os.getcwd()
        try:
            os.chdir(_dir)
            # This is roughly equivalent to:
            #    subprocess.run([sys.executable, bname],
            #                   stdout=f, stderr=f, cwd=_dir)
            with capture_output(None, True) as OUT:
                # Note: we want LoggingIntercept to log to the
                # *current* stdout (which is the TeeStream from
                # capture_output).  This ensures that log messages and
                # normal output appear in the correct order.
                with LoggingIntercept(sys.stdout, formatter=pyomo_formatter):
                    import_file(bname, infer_package=False, module_name='__main__')
        finally:
            os.chdir(cwd)

        try:
            self.compare_baseline(OUT.getvalue(), baseline)
        except:
            if os.environ.get('PYOMO_TEST_UPDATE_BASELINES', None):
                with open(base_file, 'w') as FILE:
                    FILE.write(OUT.getvalue())
            raise

    def shell_test_driver(self, tname, test_file, base_file):
        bname = os.path.basename(test_file)
        _dir = os.path.dirname(test_file)

        skip_msg = self.check_skip('test_' + tname)
        if skip_msg:
            raise _unittest.SkipTest(skip_msg)

        # Skip all shell tests on Windows.
        if os.name == 'nt':
            raise _unittest.SkipTest("Shell tests are not runnable on Windows")

        with open(base_file, 'r') as FILE:
            baseline = FILE.read()

        cwd = os.getcwd()
        try:
            os.chdir(_dir)
            _env = os.environ.copy()
            _env['PATH'] = os.pathsep.join(
                [os.path.dirname(sys.executable), _env['PATH']]
            )
            rc = subprocess.run(
                ['bash', bname],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                cwd=_dir,
                env=_env,
            )
        finally:
            os.chdir(cwd)

        try:
            self.compare_baseline(rc.stdout.decode(), baseline)
        except:
            if os.environ.get('PYOMO_TEST_UPDATE_BASELINES', None):
                with open(base_file, 'w') as FILE:
                    FILE.write(rc.stdout.decode())
            raise



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

import os
import time
import tempfile
import logging
import shutil
import weakref

from pyomo.common.dependencies import attempt_import, pyutilib_available
from pyomo.common.deprecation import deprecated, deprecation_warning
from pyomo.common.errors import TempfileContextError
from pyomo.common.multithread import MultiThreadWrapperWithMain

deletion_errors_are_fatal = True
logger = logging.getLogger(__name__)
pyutilib_tempfiles, _ = attempt_import('pyutilib.component.config.tempfiles')


class TempfileManagerClass:
    """A class for managing tempfile contexts

    Pyomo declares a global instance of this class as ``TempfileManager``:

    .. doctest::

       >>> from pyomo.common.tempfiles import TempfileManager

    This class provides an interface for managing
    :class:`TempfileContext` contexts.  It implements a basic stack,
    where users can :meth:`push()` a new context (causing it to become
    the current "active" context) and :meth:`pop()` contexts off
    (optionally deleting all files associated with the context).  In
    general usage, users will either use this class to create new
    tempfile contexts and use them explicitly (i.e., through a context
    manager):

    .. doctest::

       >>> import os
       >>> with TempfileManager.new_context() as tempfile:
       ...     fd, fname = tempfile.mkstemp()
       ...     dname = tempfile.mkdtemp()
       ...     os.path.isfile(fname)
       ...     os.path.isdir(dname)
       True
       True
       >>> os.path.exists(fname)
       False
       >>> os.path.exists(dname)
       False

    or through an implicit active context accessed through the manager
    class:

    .. doctest::

       >>> TempfileManager.push()
       <pyomo.common.tempfiles.TempfileContext object ...>
       >>> fname = TempfileManager.create_tempfile()
       >>> dname = TempfileManager.create_tempdir()
       >>> os.path.isfile(fname)
       True
       >>> os.path.isdir(dname)
       True

       >>> TempfileManager.pop()
       <pyomo.common.tempfiles.TempfileContext object ...>
       >>> os.path.exists(fname)
       False
       >>> os.path.exists(dname)
       False

    """

    def __init__(self):
        self._context_stack = []
        self._context_manager_stack = []
        self.tempdir = None

    def __del__(self):
        self.shutdown()

    def shutdown(self, remove=True):
        if not self._context_stack:
            return
        if logger is not None:
            if any(ctx.tempfiles for ctx in self._context_stack):
                logger.error(
                    "Temporary files created through TempfileManager "
                    "contexts have not been deleted (observed during "
                    "TempfileManager instance shutdown).\n"
                    "Undeleted entries:\n\t"
                    + "\n\t".join(
                        fname if isinstance(fname, str) else fname.decode()
                        for ctx in self._context_stack
                        for fd, fname in ctx.tempfiles
                    )
                )
            if self._context_stack:
                logger.warning(
                    "TempfileManagerClass instance: un-popped tempfile "
                    "contexts still exist during TempfileManager instance "
                    "shutdown"
                )
        self.clear_tempfiles(remove)
        # Delete the stack so that subsequent operations generate an
        # exception
        self._context_stack = None

    def context(self):
        """Return the current active TempfileContext.

        Raises
        ------
        TempfileContextError if there is not a current context."""
        if not self._context_stack:
            raise TempfileContextError(
                "TempfileManager has no currently active context.  "
                "Create a context (with push() or __enter__()) before "
                "attempting to create temporary objects."
            )
        return self._context_stack[-1]

    def create_tempfile(self, suffix=None, prefix=None, text=False, dir=None):
        "Call :meth:`TempfileContext.create_tempfile` on the active context"
        return self.context().create_tempfile(
            suffix=suffix, prefix=prefix, text=text, dir=dir
        )

    def create_tempdir(self, suffix=None, prefix=None, dir=None):
        "Call :meth:`TempfileContext.create_tempdir` on the active context"
        return self.context().create_tempdir(suffix=suffix, prefix=prefix, dir=dir)

    def add_tempfile(self, filename, exists=True):
        "Call :meth:`TempfileContext.add_tempfile` on the active context"
        return self.context().add_tempfile(filename=filename, exists=exists)

    def clear_tempfiles(self, remove=True):
        """Delete all temporary files and remove all contexts."""
        while self._context_stack:
            self.pop(remove)

    @deprecated(
        "The TempfileManager.sequential_files() method has been "
        "removed.  All temporary files are created with guaranteed "
        "unique names.  Users wishing sequentially numbered files "
        "should create a temporary (empty) directory using mkdtemp "
        "/ create_tempdir and place the sequential files within it.",
        version='6.2',
    )
    def sequential_files(self, ctr=0):
        pass

    def unique_files(self):
        pass

    def new_context(self):
        """Create and return an new tempfile context

        Returns
        -------
        TempfileContext
            the newly-created tempfile context

        """
        return TempfileContext(self)

    def push(self):
        """Create a new tempfile context and set it as the active context.

        Returns
        -------
        TempfileContext
            the newly-created tempfile context

        """
        context = self.new_context()
        self._context_stack.append(context)
        return context

    def pop(self, remove=True):
        """Remove and release the active context

        Parameters
        ----------
        remove: bool
            If ``True``, delete all managed files / directories

        """
        ctx = self._context_stack.pop()
        ctx.release(remove)
        return ctx

    def __enter__(self):
        ctx = self.push()
        self._context_manager_stack.append(ctx)
        return ctx

    def __exit__(self, exc_type, exc_val, exc_tb):
        ctx = self._context_manager_stack.pop()
        while True:
            if ctx is self.pop():
                break
            logger.warning(
                "TempfileManager: tempfile context was pushed onto "
                "the TempfileManager stack within a context manager "
                "(i.e., `with TempfileManager:`) but was not popped "
                "before the context manager exited.  Popping the "
                "context to preserve the stack integrity."
            )


class TempfileContext:
    """A `context` for managing collections of temporary files

    Instances of this class hold a "temporary file context".  That is,
    this records a collection of temporary file system objects that are
    all managed as a group.  The most common use of the context is to
    ensure that all files are deleted when the context is released.

    This class replicates a significant portion of the :mod:`tempfile`
    module interface.

    Instances of this class may be used as context managers (with the
    temporary files / directories getting automatically deleted when the
    context manager exits).

    Instances will also attempt to delete any temporary objects from the
    filesystem when the context falls out of scope (although this
    behavior is not guaranteed for instances existing when the
    interpreter is shutting down).

    """

    def __init__(self, manager):
        self.manager = weakref.ref(manager)
        self.tempfiles = []
        self.tempdir = None
        # Create a local reference from the TempfileContext to the os
        # and shutil modules so that this object is deleted before the
        # os and shutil modules are deallocated (since
        # TempfileContext.__del__ can call methods in those modules
        # through TempfileContext.release()).
        self.os = os
        self.shutil = shutil

    def __del__(self):
        self.release()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def mkstemp(self, suffix=None, prefix=None, dir=None, text=False, delete=True):
        """Create a unique temporary file using :func:`tempfile.mkstemp`

        Parameters are handled as in :func:`tempfile.mkstemp`, with
        the exception that the new file is created in the directory
        returned by :meth:`gettempdir`

        Returns
        -------
        fd: int
            the opened file descriptor

        fname: str or bytes
            the absolute path to the new temporary file

        """
        dir = self._resolve_tempdir(dir)
        # Note: ans == (fd, fname)
        ans = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dir, text=text)
        if delete:
            self.tempfiles.append(ans)
        return ans

    def mkdtemp(self, suffix=None, prefix=None, dir=None, delete=True):
        """Create a unique temporary directory using :func:`tempfile.mkdtemp`

        Parameters are handled as in :func:`tempfile.mkdtemp`, with
        the exception that the new file is created in the directory
        returned by :meth:`gettempdir`

        Returns
        -------
        dname: str or bytes
            the absolute path to the new temporary directory

        """
        dir = self._resolve_tempdir(dir)
        dname = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
        if delete:
            self.tempfiles.append((None, dname))
        return dname

    def gettempdir(self):
        """Return the default name of the directory used for temporary files.

        This method returns the first non-null location returned from:

         - This context's ``tempdir`` (i.e., ``self.tempdir``)
         - This context's manager's ``tempdir`` (i.e.,
           ``self.manager().tempdir``)
         - :func:`tempfile.gettempdir()`

        Returns
        -------
        dir: str
            The default directory to use for creating temporary objects
        """
        dir = self._resolve_tempdir()
        if dir is None:
            return tempfile.gettempdir()
        if isinstance(dir, bytes):
            return dir.decode()
        return dir

    def gettempdirb(self):
        """Same as :meth:`gettempdir()`, but the return value is ``bytes``"""
        dir = self._resolve_tempdir()
        if dir is None:
            return tempfile.gettempdirb()
        if not isinstance(dir, bytes):
            return dir.encode()
        return dir

    def gettempprefix(self):
        """Return the filename prefix used to create temporary files.

        See :func:`tempfile.gettempprefix()`

        """
        return tempfile.gettempprefix()

    def gettempprefixb(self):
        """Same as :meth:`gettempprefix()`, but the return value is ``bytes``"""
        return tempfile.gettempprefixb()

    def create_tempfile(self, suffix=None, prefix=None, text=False, dir=None):
        """Create a unique temporary file.

        The file name is generated as in :func:`tempfile.mkstemp()`.

        Any file handles to the new file (e.g., from :meth:`mkstemp`)
        are closed.

        Returns
        -------
        fname: str or bytes
            The absolute path of the new file.

        """
        fd, fname = self.mkstemp(suffix=suffix, prefix=prefix, dir=dir, text=text)
        os.close(fd)
        self.tempfiles[-1] = (None, fname)
        return fname

    def create_tempdir(self, suffix=None, prefix=None, dir=None):
        """Create a unique temporary directory.

        The file name is generated as in :func:`tempfile.mkdtemp()`.

        Returns
        -------
        dname: str or bytes
            The absolute path of the new directory.

        """
        return self.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)

    def add_tempfile(self, filename, exists=True):
        """Declare the specified file/directory to be temporary.

        This adds the specified path as a "temporary" object to this
        context's list of managed temporary paths (i.e., it will be
        potentially be deleted when the context is released (see
        :meth:`release`).

        Parameters
        ----------
        filename: str
            the file / directory name to be treated as temporary
        exists: bool
            if ``True``, the file / directory must already exist.

        """
        tmp = os.path.abspath(filename)
        if exists and not os.path.exists(tmp):
            raise IOError("Temporary file does not exist: " + tmp)
        self.tempfiles.append((None, tmp))

    def release(self, remove=True):
        """Release this context

        This releases the current context, potentially deleting all
        managed temporary objects (files and directories), and resetting
        the context to generate unique names.

        Parameters
        ----------
        remove: bool
            If ``True``, delete all managed files / directories
        """
        if remove and self.tempfiles:
            for fd, name in reversed(self.tempfiles):
                if fd is not None:
                    try:
                        self.os.close(fd)
                    except OSError:
                        pass
                self._remove_filesystem_object(name)
        self.tempfiles.clear()

    def _resolve_tempdir(self, dir=None):
        if dir is not None:
            return dir
        elif self.tempdir is not None:
            return self.tempdir
        elif self.manager().tempdir is not None:
            return self.manager().tempdir
        elif TempfileManager.main_thread.tempdir is not None:
            return TempfileManager.main_thread.tempdir
        elif pyutilib_available:
            if pyutilib_tempfiles.TempfileManager.tempdir is not None:
                deprecation_warning(
                    "The use of the PyUtilib TempfileManager.tempdir "
                    "to specify the default location for Pyomo "
                    "temporary files has been deprecated.  "
                    "Please set TempfileManager.tempdir in "
                    "pyomo.common.tempfiles",
                    version='5.7.2',
                )
                return pyutilib_tempfiles.TempfileManager.tempdir
        return None

    def _remove_filesystem_object(self, name):
        if not self.os.path.exists(name):
            return
        if self.os.path.isfile(name) or self.os.path.islink(name):
            try:
                self.os.remove(name)
            except WindowsError:
                # Sometimes Windows doesn't release the
                # file lock immediately when the process
                # terminates.  If we get an error, wait a
                # second and try again.
                try:
                    time.sleep(1)
                    self.os.remove(name)
                except WindowsError:
                    if deletion_errors_are_fatal:
                        raise
                    else:
                        # Failure to delete a tempfile
                        # should NOT be fatal
                        logger = logging.getLogger(__name__)
                        logger.warning("Unable to delete temporary file %s" % (name,))
            return
        assert self.os.path.isdir(name)
        self.shutil.rmtree(name, ignore_errors=not deletion_errors_are_fatal)


# The global Pyomo TempfileManager instance
TempfileManager: TempfileManagerClass = MultiThreadWrapperWithMain(TempfileManagerClass)
